# cython: profile=False
# cython: language_level=3
import logging
import os

import numpy
import pygeoprocessing
cimport numpy
cimport cython
from osgeo import gdal

from cpython.mem cimport PyMem_Malloc, PyMem_Free
from cython.operator cimport dereference as deref
from cython.operator cimport preincrement as inc
from libcpp.pair cimport pair
from libcpp.set cimport set as cset
from libcpp.list cimport list as clist
from libcpp.stack cimport stack
cimport libc.math as cmath

cdef extern from "time.h" nogil:
    ctypedef int time_t
    time_t time(time_t*)

LOGGER = logging.getLogger(__name__)


# cmath is supposed to have M_SQRT2, but tests have been failing recently
# due to a missing symbol.
cdef double SQRT2 = cmath.sqrt(2)
cdef double PI = 3.141592653589793238462643383279502884
# This module creates rasters with a memory xy block size of 2**BLOCK_BITS
cdef int BLOCK_BITS = 8
# Number of raster blocks to hold in memory at once per Managed Raster
cdef int MANAGED_RASTER_N_BLOCKS = 2**6

# These offsets are for the neighbor rows and columns according to the
# ordering: 3 2 1
#           4 x 0
#           5 6 7
cdef int *ROW_OFFSETS = [0, -1, -1, -1,  0,  1, 1, 1]
cdef int *COL_OFFSETS = [1,  1,  0, -1, -1, -1, 0, 1]


cdef int is_close(double x, double y):
    return abs(x-y) <= (1e-8+1e-05*abs(y))

# this is a least recently used cache written in C++ in an external file,
# exposing here so _ManagedRaster can use it
cdef extern from "LRUCache.h" nogil:
    cdef cppclass LRUCache[KEY_T, VAL_T]:
        LRUCache(int)
        void put(KEY_T&, VAL_T&, clist[pair[KEY_T,VAL_T]]&)
        clist[pair[KEY_T,VAL_T]].iterator begin()
        clist[pair[KEY_T,VAL_T]].iterator end()
        bint exist(KEY_T &)
        VAL_T get(KEY_T &)

# this ctype is used to store the block ID and the block buffer as one object
# inside Managed Raster
ctypedef pair[int, double*] BlockBufferPair

# a class to allow fast random per-pixel access to a raster for both setting
# and reading pixels.  Copied from src/pygeoprocessing/routing/routing.pyx,
# revision 891288683889237cfd3a3d0a1f09483c23489fca.
cdef class _ManagedRaster:
    cdef LRUCache[int, double*]* lru_cache
    cdef cset[int] dirty_blocks
    cdef int block_xsize
    cdef int block_ysize
    cdef int block_xmod
    cdef int block_ymod
    cdef int block_xbits
    cdef int block_ybits
    cdef long raster_x_size
    cdef long raster_y_size
    cdef int block_nx
    cdef int block_ny
    cdef int write_mode
    cdef bytes raster_path
    cdef int band_id
    cdef int closed

    def __cinit__(self, raster_path, band_id, write_mode):
        """Create new instance of Managed Raster.

        Args:
            raster_path (char*): path to raster that has block sizes that are
                powers of 2. If not, an exception is raised.
            band_id (int): which band in `raster_path` to index. Uses GDAL
                notation that starts at 1.
            write_mode (boolean): if true, this raster is writable and dirty
                memory blocks will be written back to the raster as blocks
                are swapped out of the cache or when the object deconstructs.

        Returns:
            None.
        """
        raster_info = pygeoprocessing.get_raster_info(raster_path)
        self.raster_x_size, self.raster_y_size = raster_info['raster_size']
        self.block_xsize, self.block_ysize = raster_info['block_size']
        self.block_xmod = self.block_xsize-1
        self.block_ymod = self.block_ysize-1

        if not (1 <= band_id <= raster_info['n_bands']):
            err_msg = (
                "Error: band ID (%s) is not a valid band number. "
                "This exception is happening in Cython, so it will cause a "
                "hard seg-fault, but it's otherwise meant to be a "
                "ValueError." % (band_id))
            print(err_msg)
            raise ValueError(err_msg)
        self.band_id = band_id

        if (self.block_xsize & (self.block_xsize - 1) != 0) or (
                self.block_ysize & (self.block_ysize - 1) != 0):
            # If inputs are not a power of two, this will at least print
            # an error message. Unfortunately with Cython, the exception will
            # present itself as a hard seg-fault, but I'm leaving the
            # ValueError in here at least for readability.
            err_msg = (
                "Error: Block size is not a power of two: "
                "block_xsize: %d, %d, %s. This exception is happening"
                "in Cython, so it will cause a hard seg-fault, but it's"
                "otherwise meant to be a ValueError." % (
                    self.block_xsize, self.block_ysize, raster_path))
            print(err_msg)
            raise ValueError(err_msg)

        self.block_xbits = numpy.log2(self.block_xsize)
        self.block_ybits = numpy.log2(self.block_ysize)
        self.block_nx = (
            self.raster_x_size + (self.block_xsize) - 1) // self.block_xsize
        self.block_ny = (
            self.raster_y_size + (self.block_ysize) - 1) // self.block_ysize

        self.lru_cache = new LRUCache[int, double*](MANAGED_RASTER_N_BLOCKS)
        self.raster_path = <bytes> raster_path
        self.write_mode = write_mode
        self.closed = 0

    def __dealloc__(self):
        """Deallocate _ManagedRaster.

        This operation manually frees memory from the LRUCache and writes any
        dirty memory blocks back to the raster if `self.write_mode` is True.
        """
        self.close()

    def close(self):
        """Close the _ManagedRaster and free up resources.

            This call writes any dirty blocks to disk, frees up the memory
            allocated as part of the cache, and frees all GDAL references.

            Any subsequent calls to any other functions in _ManagedRaster will
            have undefined behavior.
        """
        if self.closed:
            return
        self.closed = 1
        cdef int xi_copy, yi_copy
        cdef numpy.ndarray[double, ndim=2] block_array = numpy.empty(
            (self.block_ysize, self.block_xsize))
        cdef double *double_buffer
        cdef int block_xi
        cdef int block_yi
        # initially the win size is the same as the block size unless
        # we're at the edge of a raster
        cdef int win_xsize
        cdef int win_ysize

        # we need the offsets to subtract from global indexes for cached array
        cdef int xoff
        cdef int yoff

        cdef clist[BlockBufferPair].iterator it = self.lru_cache.begin()
        cdef clist[BlockBufferPair].iterator end = self.lru_cache.end()
        if not self.write_mode:
            while it != end:
                # write the changed value back if desired
                PyMem_Free(deref(it).second)
                inc(it)
            return

        raster = gdal.OpenEx(
            self.raster_path, gdal.GA_Update | gdal.OF_RASTER)
        raster_band = raster.GetRasterBand(self.band_id)

        # if we get here, we're in write_mode
        cdef cset[int].iterator dirty_itr
        while it != end:
            double_buffer = deref(it).second
            block_index = deref(it).first

            # write to disk if block is dirty
            dirty_itr = self.dirty_blocks.find(block_index)
            if dirty_itr != self.dirty_blocks.end():
                self.dirty_blocks.erase(dirty_itr)
                block_xi = block_index % self.block_nx
                block_yi = block_index / self.block_nx

                # we need the offsets to subtract from global indexes for
                # cached array
                xoff = block_xi << self.block_xbits
                yoff = block_yi << self.block_ybits

                win_xsize = self.block_xsize
                win_ysize = self.block_ysize

                # clip window sizes if necessary
                if xoff+win_xsize > self.raster_x_size:
                    win_xsize = win_xsize - (
                        xoff+win_xsize - self.raster_x_size)
                if yoff+win_ysize > self.raster_y_size:
                    win_ysize = win_ysize - (
                        yoff+win_ysize - self.raster_y_size)

                for xi_copy in xrange(win_xsize):
                    for yi_copy in xrange(win_ysize):
                        block_array[yi_copy, xi_copy] = (
                            double_buffer[
                                (yi_copy << self.block_xbits) + xi_copy])
                raster_band.WriteArray(
                    block_array[0:win_ysize, 0:win_xsize],
                    xoff=xoff, yoff=yoff)
            PyMem_Free(double_buffer)
            inc(it)
        raster_band.FlushCache()
        raster_band = None
        raster = None

    cdef inline void set(self, long xi, long yi, double value):
        """Set the pixel at `xi,yi` to `value`."""
        cdef int block_xi = xi >> self.block_xbits
        cdef int block_yi = yi >> self.block_ybits
        # this is the flat index for the block
        cdef int block_index = block_yi * self.block_nx + block_xi
        if not self.lru_cache.exist(block_index):
            self._load_block(block_index)
        self.lru_cache.get(
            block_index)[
                ((yi & (self.block_ymod))<<self.block_xbits) +
                (xi & (self.block_xmod))] = value
        if self.write_mode:
            dirty_itr = self.dirty_blocks.find(block_index)
            if dirty_itr == self.dirty_blocks.end():
                self.dirty_blocks.insert(block_index)

    cdef inline double get(self, long xi, long yi):
        """Return the value of the pixel at `xi,yi`."""
        cdef int block_xi = xi >> self.block_xbits
        cdef int block_yi = yi >> self.block_ybits
        # this is the flat index for the block
        cdef int block_index = block_yi * self.block_nx + block_xi
        if not self.lru_cache.exist(block_index):
            self._load_block(block_index)
        return self.lru_cache.get(
            block_index)[
                ((yi & (self.block_ymod))<<self.block_xbits) +
                (xi & (self.block_xmod))]

    cdef void _load_block(self, int block_index) except *:
        cdef int block_xi = block_index % self.block_nx
        cdef int block_yi = block_index // self.block_nx

        # we need the offsets to subtract from global indexes for cached array
        cdef int xoff = block_xi << self.block_xbits
        cdef int yoff = block_yi << self.block_ybits

        cdef int xi_copy, yi_copy
        cdef numpy.ndarray[double, ndim=2] block_array
        cdef double *double_buffer
        cdef clist[BlockBufferPair] removed_value_list

        # determine the block aligned xoffset for read as array

        # initially the win size is the same as the block size unless
        # we're at the edge of a raster
        cdef int win_xsize = self.block_xsize
        cdef int win_ysize = self.block_ysize

        # load a new block
        if xoff+win_xsize > self.raster_x_size:
            win_xsize = win_xsize - (xoff+win_xsize - self.raster_x_size)
        if yoff+win_ysize > self.raster_y_size:
            win_ysize = win_ysize - (yoff+win_ysize - self.raster_y_size)

        raster = gdal.OpenEx(self.raster_path, gdal.OF_RASTER)
        raster_band = raster.GetRasterBand(self.band_id)
        block_array = raster_band.ReadAsArray(
            xoff=xoff, yoff=yoff, win_xsize=win_xsize,
            win_ysize=win_ysize).astype(
            numpy.float64)
        raster_band = None
        raster = None
        double_buffer = <double*>PyMem_Malloc(
            (sizeof(double) << self.block_xbits) * win_ysize)
        for xi_copy in xrange(win_xsize):
            for yi_copy in xrange(win_ysize):
                double_buffer[(yi_copy<<self.block_xbits)+xi_copy] = (
                    block_array[yi_copy, xi_copy])
        self.lru_cache.put(
            <int>block_index, <double*>double_buffer, removed_value_list)

        if self.write_mode:
            raster = gdal.OpenEx(
                self.raster_path, gdal.GA_Update | gdal.OF_RASTER)
            raster_band = raster.GetRasterBand(self.band_id)

        block_array = numpy.empty(
            (self.block_ysize, self.block_xsize), dtype=numpy.double)
        while not removed_value_list.empty():
            # write the changed value back if desired
            double_buffer = removed_value_list.front().second

            if self.write_mode:
                block_index = removed_value_list.front().first

                # write back the block if it's dirty
                dirty_itr = self.dirty_blocks.find(block_index)
                if dirty_itr != self.dirty_blocks.end():
                    self.dirty_blocks.erase(dirty_itr)

                    block_xi = block_index % self.block_nx
                    block_yi = block_index // self.block_nx

                    xoff = block_xi << self.block_xbits
                    yoff = block_yi << self.block_ybits

                    win_xsize = self.block_xsize
                    win_ysize = self.block_ysize

                    if xoff+win_xsize > self.raster_x_size:
                        win_xsize = win_xsize - (
                            xoff+win_xsize - self.raster_x_size)
                    if yoff+win_ysize > self.raster_y_size:
                        win_ysize = win_ysize - (
                            yoff+win_ysize - self.raster_y_size)

                    for xi_copy in xrange(win_xsize):
                        for yi_copy in xrange(win_ysize):
                            block_array[yi_copy, xi_copy] = double_buffer[
                                (yi_copy << self.block_xbits) + xi_copy]
                    raster_band.WriteArray(
                        block_array[0:win_ysize, 0:win_xsize],
                        xoff=xoff, yoff=yoff)
            PyMem_Free(double_buffer)
            removed_value_list.pop_front()

        if self.write_mode:
            raster_band = None
            raster = None


def calculate_sediment_deposition(
        mfd_flow_direction_path, e_prime_path, f_path, sdr_path,
        target_sediment_deposition_path):
    """Calculate sediment deposition layer.

    This algorithm outputs both sediment deposition (r_i) and flux (f_i)::

        r_i  =      dr_i  * (sum over j ∈ J of f_j * p(i,j)) + E'_i

        f_i  = (1 - dr_i) * (sum over j ∈ J of f_j * p(i,j)) + E'_i


                (sum over k ∈ K of SDR_k * p(i,k)) - SDR_i
        dr_i = --------------------------------------------
                              (1 - SDR_i)

    where:

    - ``p(i,j)`` is the proportion of flow from pixel ``i`` into pixel ``j``
    - ``J`` is the set of pixels that are immediate upstream neighbors of
      pixel ``i``
    - ``K`` is the set of pixels that are immediate downstream neighbors of
      pixel ``i``
    - ``E'`` is ``USLE * (1 - SDR)``, the amount of sediment loss from pixel
      ``i`` that doesn't reach a stream (``e_prime_path``)
    - ``SDR`` is the sediment delivery ratio (``sdr_path``)

    ``f_i`` is recursively defined in terms of ``i``'s upstream neighbors.
    The algorithm begins from seed pixels that are local high points and so
    have no upstream neighbors. It works downstream from each seed pixel,
    only adding a pixel to the stack when all its upstream neighbors are
    already calculated.

    Note that this function is designed to be used in the context of the SDR
    model. Because the algorithm is recursive upstream and downstream of each
    pixel, nodata values in the SDR input would propagate along the flow path.
    This case is not handled because we assume the SDR and flow dir inputs
    will come from the SDR model and have nodata in the same places.

    Args:
        mfd_flow_direction_path (string): a path to a raster with
            pygeoprocessing.routing MFD flow direction values.
        e_prime_path (string): path to a raster that shows sources of
            sediment that wash off a pixel but do not reach the stream.
        f_path (string): path to a raster that shows the sediment flux
            on a pixel for sediment that does not reach the stream.
        sdr_path (string): path to Sediment Delivery Ratio raster.
        target_sediment_deposition_path (string): path to created that
            shows where the E' sources end up across the landscape.

    Returns:
        None.

    """
    LOGGER.info('Calculate sediment deposition')
    cdef float target_nodata = -1
    pygeoprocessing.new_raster_from_base(
        mfd_flow_direction_path, target_sediment_deposition_path,
        gdal.GDT_Float32, [target_nodata])
    pygeoprocessing.new_raster_from_base(
        mfd_flow_direction_path, f_path,
        gdal.GDT_Float32, [target_nodata])

    cdef _ManagedRaster mfd_flow_direction_raster = _ManagedRaster(
        mfd_flow_direction_path, 1, False)
    cdef _ManagedRaster e_prime_raster = _ManagedRaster(
        e_prime_path, 1, False)
    cdef _ManagedRaster sdr_raster = _ManagedRaster(sdr_path, 1, False)
    cdef _ManagedRaster f_raster = _ManagedRaster(f_path, 1, True)
    cdef _ManagedRaster sediment_deposition_raster = _ManagedRaster(
        target_sediment_deposition_path, 1, True)

    # given the pixel neighbor numbering system
    #  3 2 1
    #  4 x 0
    #  5 6 7
    # if a pixel `x` has a neighbor `n` in position `i`,
    # then `n`'s neighbor in position `inflow_offsets[i]`
    # is the original pixel `x`
    cdef int *inflow_offsets = [4, 5, 6, 7, 0, 1, 2, 3]

    cdef int n_cols, n_rows
    flow_dir_info = pygeoprocessing.get_raster_info(mfd_flow_direction_path)
    n_cols, n_rows = flow_dir_info['raster_size']
    cdef int mfd_nodata = 0
    cdef stack[int] processing_stack
    cdef float sdr_nodata = pygeoprocessing.get_raster_info(
        sdr_path)['nodata'][0]
    cdef float e_prime_nodata = pygeoprocessing.get_raster_info(
        e_prime_path)['nodata'][0]
    cdef int col_index, row_index, win_xsize, win_ysize, xoff, yoff
    cdef int global_col, global_row, flat_index, j, k
    cdef int seed_col = 0
    cdef int seed_row = 0
    cdef int neighbor_row, neighbor_col
    cdef int flow_val, neighbor_flow_val, ds_neighbor_flow_val
    cdef int flow_weight, neighbor_flow_weight
    cdef float flow_sum, neighbor_flow_sum
    cdef float downstream_sdr_weighted_sum, sdr_i, sdr_j
    cdef float p_j, p_val

    for offset_dict in pygeoprocessing.iterblocks(
            (mfd_flow_direction_path, 1), offset_only=True, largest_block=0):
        win_xsize = offset_dict['win_xsize']
        win_ysize = offset_dict['win_ysize']
        xoff = offset_dict['xoff']
        yoff = offset_dict['yoff']

        LOGGER.info('Sediment deposition %.2f%% complete', 100 * (
            (xoff * yoff) / float(n_cols*n_rows)))

        for row_index in range(win_ysize):
            seed_row = yoff + row_index
            for col_index in range(win_xsize):
                seed_col = xoff + col_index
                # check if this is a good seed pixel ( a local high point)
                if mfd_flow_direction_raster.get(seed_col, seed_row) == mfd_nodata:
                    continue
                seed_pixel = 1
                # iterate over each of the pixel's neighbors
                for j in range(8):
                    # skip if the neighbor is outside the raster bounds
                    neighbor_row = seed_row + ROW_OFFSETS[j]
                    if neighbor_row < 0 or neighbor_row >= n_rows:
                        continue
                    neighbor_col = seed_col + COL_OFFSETS[j]
                    if neighbor_col < 0 or neighbor_col >= n_cols:
                        continue
                    # skip if the neighbor's flow direction is undefined
                    neighbor_flow_val = <int>mfd_flow_direction_raster.get(
                        neighbor_col, neighbor_row)
                    if neighbor_flow_val == mfd_nodata:
                        continue
                    # if the neighbor flows into it, it's not a local high
                    # point and so can't be a seed pixel
                    neighbor_flow_weight = (
                        neighbor_flow_val >> (inflow_offsets[j]*4)) & 0xF
                    if neighbor_flow_weight > 0:
                        seed_pixel = 0  # neighbor flows in, not a seed
                        break

                # if this can be a seed pixel and hasn't already been
                # calculated, put it on the stack
                if seed_pixel and sediment_deposition_raster.get(
                        seed_col, seed_row) == target_nodata:
                    processing_stack.push(seed_row * n_cols + seed_col)

                while processing_stack.size() > 0:
                    # loop invariant: cell has all upstream neighbors
                    # processed. this is true for seed pixels because they
                    # have no upstream neighbors.
                    flat_index = processing_stack.top()
                    processing_stack.pop()
                    global_row = flat_index // n_cols
                    global_col = flat_index % n_cols

                    # (sum over j ∈ J of f_j * p(i,j) in the equation for r_i)
                    # calculate the upstream f_j contribution to this pixel,
                    # the weighted sum of flux flowing onto this pixel from
                    # all neighbors
                    f_j_weighted_sum = 0
                    for j in range(8):
                        neighbor_row = global_row + ROW_OFFSETS[j]
                        if neighbor_row < 0 or neighbor_row >= n_rows:
                            continue
                        neighbor_col = global_col + COL_OFFSETS[j]
                        if neighbor_col < 0 or neighbor_col >= n_cols:
                            continue

                        # see if there's an inflow from the neighbor to the
                        # pixel
                        neighbor_flow_val = (
                            <int>mfd_flow_direction_raster.get(
                                neighbor_col, neighbor_row))
                        neighbor_flow_weight = (
                            neighbor_flow_val >> (inflow_offsets[j]*4)) & 0xF
                        if neighbor_flow_weight > 0:
                            f_j = f_raster.get(neighbor_col, neighbor_row)
                            if f_j == target_nodata:
                                continue
                            # sum up the neighbor's flow dir values in each
                            # direction.
                            # flow dir values are relative to the total
                            neighbor_flow_sum = 0
                            for k in range(8):
                                neighbor_flow_sum += (
                                    neighbor_flow_val >> (k*4)) & 0xF
                            # get the proportion of the neighbor's flow that
                            # flows into the original pixel
                            p_val = neighbor_flow_weight / neighbor_flow_sum
                            # add the neighbor's flux value, weighted by the 
                            # flow proportion
                            f_j_weighted_sum += p_val * f_j

                    # calculate sum of SDR values of immediate downstream 
                    # neighbors, weighted by proportion of flow into each
                    # neighbor
                    # (sum over k ∈ K of SDR_k * p(i,k) in the equation above)
                    downstream_sdr_weighted_sum = 0
                    flow_val = <int>mfd_flow_direction_raster.get(
                        global_col, global_row)
                    flow_sum = 0
                    for k in range(8):
                        flow_sum += (flow_val >> (k*4)) & 0xF

                    # iterate over the neighbors again
                    for j in range(8):
                        # skip if neighbor is outside the raster boundaries
                        neighbor_row = global_row + ROW_OFFSETS[j]
                        if neighbor_row < 0 or neighbor_row >= n_rows:
                            continue
                        neighbor_col = global_col + COL_OFFSETS[j]
                        if neighbor_col < 0 or neighbor_col >= n_cols:
                            continue
                        # if it is a downstream neighbor, add to the sum and
                        # check if it can be pushed onto the stack yet
                        flow_weight = (flow_val >> (j*4)) & 0xF
                        if flow_weight > 0:
                            sdr_j = sdr_raster.get(neighbor_col, neighbor_row)
                            if sdr_j == sdr_nodata:
                                continue
                            if sdr_j == 0:
                                # this means it's a stream, for SDR deposition
                                # purposes, we set sdr to 1 to indicate this
                                # is the last step on which to retain sediment
                                sdr_j = 1
                            p_j = flow_weight / flow_sum
                            downstream_sdr_weighted_sum += sdr_j * p_j

                            # check if we can add neighbor j to the stack yet
                            #
                            # if there is a downstream neighbor it
                            # couldn't have been pushed on the processing
                            # stack yet, because the upstream was just
                            # completed
                            upstream_neighbors_processed = 1
                            # iterate over each neighbor-of-neighbor
                            for k in range(8):
                                # no need to push the one we're currently
                                # calculating back onto the stack
                                if inflow_offsets[k] == j:
                                    continue
                                # skip if neighbor-of-neighbor is outside
                                # raster bounds
                                ds_neighbor_row = (
                                    neighbor_row + ROW_OFFSETS[k])
                                if ds_neighbor_row < 0 or ds_neighbor_row >= n_rows:
                                    continue
                                ds_neighbor_col = (
                                    neighbor_col + COL_OFFSETS[k])
                                if ds_neighbor_col < 0 or ds_neighbor_col >= n_cols:
                                    continue
                                # if any upstream neighbor of j hasn't been
                                # calculated, we can't push j onto the stack
                                # yet
                                ds_neighbor_flow_val = (
                                    <int>mfd_flow_direction_raster.get(
                                        ds_neighbor_col, ds_neighbor_row))
                                if (ds_neighbor_flow_val >> (
                                        inflow_offsets[k]*4)) & 0xF > 0:
                                    if (sediment_deposition_raster.get(
                                            ds_neighbor_col, ds_neighbor_row) ==
                                            target_nodata):
                                        upstream_neighbors_processed = 0
                                        break
                            # if all upstream neighbors of neighbor j are
                            # processed, we can push j onto the stack.
                            if upstream_neighbors_processed:
                                processing_stack.push(
                                    neighbor_row * n_cols +
                                    neighbor_col)

                    # nodata pixels should propagate to the results
                    sdr_i = sdr_raster.get(global_col, global_row)
                    if sdr_i == sdr_nodata:
                        continue
                    e_prime_i = e_prime_raster.get(global_col, global_row)
                    if e_prime_i == e_prime_nodata:
                        continue

                    if downstream_sdr_weighted_sum < sdr_i:
                        # i think this happens because of our low resolution
                        # flow direction, it's okay to zero out.
                        downstream_sdr_weighted_sum = sdr_i

                    # these correspond to the full equations for
                    # dr_i, r_i, and f_i given in the docstring
                    dr_i = (downstream_sdr_weighted_sum - sdr_i) / (1 - sdr_i)
                    r_i = dr_i * (e_prime_i + f_j_weighted_sum)
                    f_i = (1 - dr_i) * (e_prime_i + f_j_weighted_sum)

                    sediment_deposition_raster.set(global_col, global_row, r_i)
                    f_raster.set(global_col, global_row, f_i)

    LOGGER.info('Sediment deposition 100% complete')
    sediment_deposition_raster.close()


def calculate_average_aspect(
        mfd_flow_direction_path, target_average_aspect_path):
    """Calculate the Weighted Average Aspect Ratio from MFD.

    Calculates the average aspect ratio weighted by proportional flow
    direction.

    Args:
        mfd_flow_direction_path (string): The path to an MFD flow direction
            raster.
        target_average_aspect_path (string): The path to where the calculated
            weighted average aspect raster should be written.

    Returns:
        ``None``.

    """
    LOGGER.info('Calculating average aspect')

    cdef float average_aspect_nodata = -1
    pygeoprocessing.new_raster_from_base(
        mfd_flow_direction_path, target_average_aspect_path,
        gdal.GDT_Float32, [average_aspect_nodata], [average_aspect_nodata])

    flow_direction_info = pygeoprocessing.get_raster_info(
        mfd_flow_direction_path)
    cdef int mfd_flow_direction_nodata = flow_direction_info['nodata'][0]
    cdef int n_cols, n_rows
    n_cols, n_rows = flow_direction_info['raster_size']

    cdef _ManagedRaster mfd_flow_direction_raster = _ManagedRaster(
        mfd_flow_direction_path, 1, False)

    cdef _ManagedRaster average_aspect_raster = _ManagedRaster(
        target_average_aspect_path, 1, True)

    cdef int seed_row = 0
    cdef int seed_col = 0
    cdef int n_pixels_visited = 0
    cdef int win_xsize, win_ysize, xoff, yoff
    cdef int row_index, col_index, neighbor_index
    cdef int flow_weight_in_direction
    cdef int weight_sum
    cdef int seed_flow_value
    cdef float aspect_weighted_average, aspect_weighted_sum

    # the flow_lengths array is the functional equivalent
    # of calculating |sin(alpha)| + |cos(alpha)|.
    cdef float* flow_lengths = [
        1, <float>SQRT2,
        1, <float>SQRT2,
        1, <float>SQRT2,
        1, <float>SQRT2
    ]

    # Loop over iterblocks to maintain cache locality
    # Find each non-nodata pixel and calculate proportional flow
    # Multiply proportional flow times the flow length x_d
    # write the final value to the raster.
    for offset_dict in pygeoprocessing.iterblocks(
            (mfd_flow_direction_path, 1), offset_only=True, largest_block=0):
        win_xsize = offset_dict['win_xsize']
        win_ysize = offset_dict['win_ysize']
        xoff = offset_dict['xoff']
        yoff = offset_dict['yoff']

        LOGGER.info('Average aspect %.2f%% complete', 100 * (
            n_pixels_visited / float(n_cols * n_rows)))

        for row_index in range(win_ysize):
            seed_row = yoff + row_index
            for col_index in range(win_xsize):
                seed_col = xoff + col_index
                seed_flow_value = <int>mfd_flow_direction_raster.get(
                    seed_col, seed_row)

                # Skip this seed if it's nodata (Currently expected to be 0).
                # No need to set the nodata value here since we have already
                # filled the raster with nodata values at creation time.
                if seed_flow_value == mfd_flow_direction_nodata:
                    continue

                weight_sum = 0
                aspect_weighted_sum = 0
                for neighbor_index in range(8):
                    neighbor_row = seed_row + ROW_OFFSETS[neighbor_index]
                    if neighbor_row == -1 or neighbor_row == n_rows:
                        continue

                    neighbor_col = seed_col + COL_OFFSETS[neighbor_index]
                    if neighbor_col == -1 or neighbor_col == n_cols:
                        continue

                    flow_weight_in_direction = (seed_flow_value >> (
                        neighbor_index * 4) & 0xF)
                    weight_sum += flow_weight_in_direction

                    aspect_weighted_sum += (
                        flow_lengths[neighbor_index] *
                        flow_weight_in_direction)

                # Weight sum should never be less than 0.
                # Since it's an int, we can compare it directly against the
                # value of 0.
                if weight_sum == 0:
                    aspect_weighted_average = average_aspect_nodata
                else:
                    # We already know that weight_sum will be > 0 because we
                    # check for it in the condition above.
                    with cython.cdivision(True):
                        aspect_weighted_average = (
                            aspect_weighted_sum / <float>weight_sum)

                average_aspect_raster.set(
                    seed_col, seed_row, aspect_weighted_average)

        n_pixels_visited += win_xsize * win_ysize

    LOGGER.info('Average aspect 100.00% complete')

    mfd_flow_direction_raster.close()
    average_aspect_raster.close()



"""Buffered file manager module."""

import uuid
import time
import collections
import os
import logging
import sqlite3

import numpy

from .. import utils


LOGGER = logging.getLogger(
    'natcap.invest.recmodel_server.buffered_numpy_disk_map')


class BufferedNumpyDiskMap(object):
    """Persistent object to append and read numpy arrays to unique keys.

    This object is abstractly a key/value pair map where the operations are
    to append, read, and delete numpy arrays associated with those keys.  The
    object attempts to keep data in RAM as much as possible and saves data to
    files on disk to manage memory and persist between instantiations.
    """

    _ARRAY_TUPLE_TYPE = numpy.dtype('datetime64[D],a4,f4,f4')

    def __init__(self, manager_filename, max_bytes_to_buffer):
        """Create file manager object.

        Args:
            manager_filename (string): path to store file manager database.
                Additional files will be created in this directory to store
                binary data as needed.
            max_bytes_to_buffer (int): number of bytes to hold in memory at
                one time.

        Returns:
            None
        """
        self.manager_filename = manager_filename
        self.manager_directory = os.path.dirname(manager_filename)
        utils.make_directories([self.manager_directory])
        db_connection = sqlite3.connect(
            manager_filename, detect_types=sqlite3.PARSE_DECLTYPES)
        db_cursor = db_connection.cursor()
        db_cursor.execute("""CREATE TABLE IF NOT EXISTS array_table
            (array_id INTEGER PRIMARY KEY, array_path TEXT)""")

        db_connection.commit()
        db_connection.close()

        self.array_cache = collections.defaultdict(collections.deque)
        self.max_bytes_to_buffer = max_bytes_to_buffer
        self.current_bytes_in_system = 0

    def append(self, array_id, array_data):
        """Append data to the file.

        Args:
            array_id (int): unique key to identify the array node
            array_data (numpy.ndarray): data to append to node.

        Returns:
            None
        """
        self.array_cache[array_id].append(array_data.copy())
        self.current_bytes_in_system += (
            array_data.size * BufferedNumpyDiskMap._ARRAY_TUPLE_TYPE.itemsize)
        if self.current_bytes_in_system > self.max_bytes_to_buffer:
            self.flush()

    def flush(self):
        """Method to flush data in memory to disk."""
        start_time = time.time()
        LOGGER.info(
            'Flushing %d bytes in %d arrays', self.current_bytes_in_system,
            len(self.array_cache))

        db_connection = sqlite3.connect(
            self.manager_filename, detect_types=sqlite3.PARSE_DECLTYPES)
        db_cursor = db_connection.cursor()

        # get all the array data to append at once
        insert_list = []

        while len(self.array_cache) > 0:
            array_id = next(iter(self.array_cache.keys()))
            array_deque = self.array_cache.pop(array_id)
            # try to get data if it's there
            db_cursor.execute(
                """SELECT (array_path) FROM array_table
                    where array_id=? LIMIT 1""", [array_id])
            array_path = db_cursor.fetchone()
            if array_path is not None:
                # cache gets wiped at end so okay to use same deque
                array_deque.append(numpy.load(array_path[0]))
                array_data = numpy.concatenate(array_deque)
                array_deque = None
                numpy.save(array_path[0], array_data)
            else:
                # otherwise directly write
                # make a random filename and put it one directory deep named
                # off the last two characters in the filename
                array_filename = uuid.uuid4().hex + '.npy'
                # -6:-4 skips the extension and gets the last 2 characters
                array_directory = os.path.join(
                    self.manager_directory, array_filename[-6:-4])
                if not os.path.isdir(array_directory):
                    os.mkdir(array_directory)
                array_path = os.path.join(array_directory, array_filename)
                # save the file
                array_data = numpy.concatenate(array_deque)
                array_deque = None
                numpy.save(array_path, array_data)
                insert_list.append((array_id, array_path))
        db_cursor.executemany(
            """INSERT INTO array_table
                (array_id, array_path)
            VALUES (?,?)""", insert_list)

        db_connection.commit()
        db_connection.close()

        self.current_bytes_in_system = 0
        LOGGER.info('Completed flush in %.2fs', time.time() - start_time)

    def read(self, array_id):
        """Read the entirety of the file.

        Internally this might mean that part of the file is read from disk
        and the end from the buffer or any combination of those.

        Args:
            array_id (string): unique node id to read

        Returns:
            contents of node as a numpy.ndarray.
        """
        db_connection = sqlite3.connect(
            self.manager_filename, detect_types=sqlite3.PARSE_DECLTYPES)
        db_cursor = db_connection.cursor()
        db_cursor.execute(
            "SELECT (array_path) FROM array_table where array_id=? LIMIT 1",
            [array_id])
        array_path = db_cursor.fetchone()
        if array_path is not None:
            array_data = numpy.load(array_path[0])
        else:
            array_data = numpy.empty(
                0, dtype=BufferedNumpyDiskMap._ARRAY_TUPLE_TYPE)

        if len(self.array_cache[array_id]) > 0:
            local_deque = collections.deque(self.array_cache[array_id])
            local_deque.append(array_data)
            array_data = numpy.concatenate(local_deque)

        return array_data

    def delete(self, array_id):
        """Delete node `array_id` from disk and cache."""
        db_connection = sqlite3.connect(
            self.manager_filename, detect_types=sqlite3.PARSE_DECLTYPES)
        db_cursor = db_connection.cursor()
        db_cursor.execute(
            "SELECT (array_path) FROM array_table where array_id=? LIMIT 1",
            [array_id])
        array_path = db_cursor.fetchone()
        if array_path is not None:
            os.remove(array_path[0])
            try:
                # attempt to remove the directory if it's empty
                os.rmdir(os.path.dirname(array_path[0]))
            except OSError:
                # it's not empty, not a big deal
                pass

        # delete the key from the table
        db_cursor.execute(
            "DELETE FROM array_table where array_id=?", [array_id])
        db_connection.close()

        # delete the cache and update cache size
        # The * 12 comes from the fact that the array is an 'a4 f4 f4'
        self.current_bytes_in_system -= (
            sum([x.size for x in self.array_cache[array_id]]) *
            BufferedNumpyDiskMap._ARRAY_TUPLE_TYPE.itemsize)
        del self.array_cache[array_id]



# coding=UTF-8

from natcap.invest.ui import model, inputs
from natcap.invest import carbon, MODEL_METADATA


class Carbon(model.InVESTModel):
    def __init__(self):
        model.InVESTModel.__init__(
            self,
            label=MODEL_METADATA['carbon'].model_title,
            target=carbon.execute,
            validator=carbon.validate,
            localdoc=MODEL_METADATA['carbon'].userguide)

        self.cur_lulc_raster = inputs.File(
            args_key='lulc_cur_path',
            helptext=(
                "A GDAL-supported raster representing the land-cover "
                "of the current scenario."),
            label='Current Land Use/Land Cover (Raster)',
            validator=self.validator)
        self.add_input(self.cur_lulc_raster)
        self.carbon_pools_path = inputs.File(
            args_key='carbon_pools_path',
            helptext=(
                "A table that maps the land-cover IDs to carbon "
                "pools.  The table must contain columns of 'LULC', "
                "'C_above', 'C_Below', 'C_Soil', 'C_Dead' as described "
                "in the User's Guide.  The values in LULC must at "
                "least include the LULC IDs in the land cover maps."),
            label='Carbon Pools',
            validator=self.validator)
        self.add_input(self.carbon_pools_path)
        self.calc_sequestration = inputs.Checkbox(
            helptext=(
                "Check to enable sequestration analysis. This "
                "requires inputs of Land Use/Land Cover maps for both "
                "current and future scenarios."),
            args_key='calc_sequestration',
            label='Calculate Sequestration')
        self.add_input(self.calc_sequestration)
        self.fut_lulc_raster = inputs.File(
            args_key='lulc_fut_path',
            helptext=(
                "A GDAL-supported raster representing the land-cover "
                "of the future scenario.  <br><br>If REDD scenario "
                "analysis is enabled, this should be the reference, or "
                "baseline, future scenario against which to compare "
                "the REDD policy scenario."),
            interactive=False,
            label='Future Land Cover (Raster)',
            validator=self.validator)
        self.add_input(self.fut_lulc_raster)
        self.redd = inputs.Checkbox(
            helptext=(
                "Check to enable REDD scenario analysis.  This "
                "requires three Land Use/Land Cover maps: one for the "
                "current scenario, one for the future baseline "
                "scenario, and one for the future REDD policy "
                "scenario."),
            interactive=False,
            args_key='do_redd',
            label='REDD Scenario Analysis')
        self.add_input(self.redd)
        self.redd_lulc_raster = inputs.File(
            args_key='lulc_redd_path',
            helptext=(
                "A GDAL-supported raster representing the land-cover "
                "of the REDD policy future scenario.  This scenario "
                "will be compared to the baseline future scenario."),
            interactive=False,
            label='REDD Policy (Raster)',
            validator=self.validator)
        self.add_input(self.redd_lulc_raster)
        self.valuation_container = inputs.Container(
            args_key='do_valuation',
            expandable=True,
            expanded=False,
            interactive=False,
            label='Run Valuation Model')
        self.add_input(self.valuation_container)
        self.cur_lulc_year = inputs.Text(
            args_key='lulc_cur_year',
            helptext='The calendar year of the current scenario.',
            interactive=False,
            label='Current Land Cover Calendar Year',
            validator=self.validator)
        self.valuation_container.add_input(self.cur_lulc_year)
        self.fut_lulc_year = inputs.Text(
            args_key='lulc_fut_year',
            helptext='The calendar year of the future scenario.',
            interactive=False,
            label='Future Land Cover Calendar Year',
            validator=self.validator)
        self.valuation_container.add_input(self.fut_lulc_year)
        self.price_per_metric_ton_of_c = inputs.Text(
            args_key='price_per_metric_ton_of_c',
            label='Price/Metric Ton of Carbon',
            validator=self.validator)
        self.valuation_container.add_input(self.price_per_metric_ton_of_c)
        self.discount_rate = inputs.Text(
            args_key='discount_rate',
            helptext='The discount rate as a floating point percent.',
            label='Market Discount in Price of Carbon (%)',
            validator=self.validator)
        self.valuation_container.add_input(self.discount_rate)
        self.rate_change = inputs.Text(
            args_key='rate_change',
            helptext=(
                "The floating point percent increase of the price of "
                "carbon per year."),
            label='Annual Rate of Change in Price of Carbon (%)',
            validator=self.validator)
        self.valuation_container.add_input(self.rate_change)

        # Set interactivity, requirement as input sufficiency changes
        self.calc_sequestration.sufficiency_changed.connect(
            self.fut_lulc_raster.set_interactive)
        self.calc_sequestration.sufficiency_changed.connect(
            self.redd.set_interactive)
        self.redd.sufficiency_changed.connect(
            self.redd_lulc_raster.set_interactive)
        self.calc_sequestration.sufficiency_changed.connect(
            self.valuation_container.set_interactive)

    def assemble_args(self):
        args = {
            self.workspace.args_key: self.workspace.value(),
            self.suffix.args_key: self.suffix.value(),
            self.cur_lulc_raster.args_key: self.cur_lulc_raster.value(),
            self.carbon_pools_path.args_key: self.carbon_pools_path.value(),
            self.valuation_container.args_key:
                self.valuation_container.value(),
            self.calc_sequestration.args_key:
                self.calc_sequestration.value(),
            self.redd.args_key: self.redd.value(),
        }

        if self.calc_sequestration.value():
            args[self.redd_lulc_raster.args_key] = (
                self.redd_lulc_raster.value())
            args[self.fut_lulc_raster.args_key] = self.fut_lulc_raster.value()

            for arg in (self.cur_lulc_year, self.fut_lulc_year):
                args[arg.args_key] = arg.value()

            # Attempt to cast valuation parameters to float
            if self.valuation_container.value():
                for arg in (self.price_per_metric_ton_of_c,
                            self.discount_rate, self.rate_change):
                    args[arg.args_key] = arg.value()

        return args



"""GLOBIO InVEST Model."""
import os
import logging
import collections
import tempfile

from osgeo import gdal
from osgeo import ogr
from osgeo import osr
import numpy
import pygeoprocessing
import taskgraph

from . import utils
from . import spec_utils
from .spec_utils import u
from . import validation
from . import MODEL_METADATA


LOGGER = logging.getLogger(__name__)

# this value of sigma == 9.0 was derived by Justin Johnson as a good
# approximation to use as a gaussian filter to replace the connectivity index.
# I don't have any other documentation than his original code base.
SIGMA = 9.0

ARGS_SPEC = {
    "model_name": MODEL_METADATA["globio"].model_title,
    "pyname": MODEL_METADATA["globio"].pyname,
    "userguide_html": MODEL_METADATA["globio"].userguide,
    "args_with_spatial_overlap": {
        "spatial_keys": [
            "lulc_path", "pasture_path", "potential_vegetation_path",
            "aoi_path", "globio_lulc_path"],
    },
    "args": {
        "workspace_dir": spec_utils.WORKSPACE,
        "results_suffix": spec_utils.SUFFIX,
        "n_workers": spec_utils.N_WORKERS,
        "predefined_globio": {
            "type": "boolean",
            "required": False,
            "about": _(
                "Use the predefined GLOBIO LULC map instead of providing "
                "the LULC map."),
            "name": _("use predefined LULC")
        },
        "lulc_path": {
            **spec_utils.LULC,
            "projected": True,
            "required": "not predefined_globio",
            "about": _(
                f"{spec_utils.LULC['about']} Each LULC code must have a "
                "corresponding entry in the biophysical table. "
                "Required if Use Predefined GLOBIO LULC is not selected.")
        },
        "lulc_to_globio_table_path": {
            "type": "csv",
            "columns": {
                "lucode": {
                    "type": "integer",
                    "about": _("LULC code from the LULC map input.")},
                "globio_lucode": {
                    "type": "integer",
                    "about": _("Corresponding GLOBIO LULC code.")}
            },
            "required": "not predefined_globio",
            "about": _(
                "A table mapping each LULC code in the LULC raster input to "
                "the corresponding GLOBIO LULC code for the class. "
                "Required if Use Predefined GLOBIO LULC is not selected."),
            "name": _("LULC to GLOBIO LULC Table")
        },
        "infrastructure_dir": {
            "type": "directory",
            "contents": {
                "[INFRASTRUCTURE_MAP]": {  # may be named anything
                    "about": _(
                        "Maps of the location of infrastructure. For rasters, "
                        "any valid value greater than zero indicates the "
                        "presence of infrastrucutre. For vectors, any "
                        "geometries indicate the presence of infrastructure."),
                    "type": {"raster", "vector"},
                    "bands": {1: {"type": "number", "units": u.none}},
                    "fields": {},
                    "geometries": spec_utils.ALL_GEOMS
                }
            },
            "about": _(
                "Directory containing raster and/or vector map(s) any forms "
                "of infrastructure to consider in the MSA calculation."),
            "name": _("infrastructure directory")
        },
        "pasture_path": {
            "type": "raster",
            "bands": {1: {"type": "ratio"}},
            "projected": True,
            "required": "not predefined_globio",
            "about": _(
                "Map of the proportion of each pixel that is pasture. "
                "Required if Use Predefined GLOBIO LULC is not selected."),
            "name": _("pasture")
        },
        "potential_vegetation_path": {
            "name": _("potential vegetation"),
            "type": "raster",
            "bands": {1: {"type": "integer"}},
            "projected": True,
            "required": "not predefined_globio",
            "about": _(
                "Map of potential vegetation classes from Ramankutty "
                "and Foley (1999). "
                "Required if Use Predefined GLOBIO LULC is not selected.")
        },
        "pasture_threshold": {
            "type": "ratio",
            "required": "not predefined_globio",
            "about": _(
                "Areas in the Pasture map with a pasture proportion greater "
                "than or equal to this threshold are considered grassland or "
                "livestock grazing areas. "
                "Required if Use Predefined GLOBIO LULC is not selected."),
            "name": _("pasture threshold")
        },
        "intensification_fraction": {
            "type": "ratio",
            "about": _(
                "Proportion of total agriculture that is intensive. "),
            "name": _("proportion of intensified agriculture")
        },
        "primary_threshold": {
            "type": "ratio",
            "required": "not predefined_globio",
            "about": _(
                "Forest ragmentation quality index threshold below which an "
                "area is classified as secondary forest. Areas with FFQI "
                "greater than or equal to this threshold are classified as "
                "primary forest. "
                "Required if Use Predefined GLOBIO LULC is not selected."),
            "name": _("primary threshold")
        },
        "msa_parameters_path": {
            "type": "csv",
            "columns": {
                "msa_type": {
                    "type": "option_string",
                    "options": {
                        "msa_i_primary": {
                            "description": _(
                                "This MSA value represents infrastructure "
                                "impacts on primary vegetation. The value in "
                                "the 'value' column is a distance range in "
                                "meters.")},
                        "msa_i_other": {
                            "description": _(
                                "This MSA value represents infrastructure "
                                "impacts on non-primary vegetation. The value "
                                "in the 'value' column is a distance range in "
                                "meters.")},
                        "msa_f": {
                            "description": _(
                                "This MSA value represents fragmentation "
                                "impacts. The value in the 'value' column is "
                                "an FFQI range.")},
                        "msa_lu": {
                            "description": _(
                                "This MSA value represents land-use impacts. "
                                "The value in the 'value' column is one of "
                                "the GLOBIO-recognized LULC codes.")},
                    },
                    "about": _("The type of MSA value in this row.")
                },
                "value": {
                    "type": "freestyle_string",
                    "about": _(
                        "Indicates a number or range of a bin. This may be a "
                        "single number e.g. 1000, a range (two numbers "
                        "separated by a hyphen e.g. 1000-2000), or an upper "
                        "or lower bound (a number preceded by > or < e.g. <5")
                },
                "msa_x": {
                    "type": "ratio",
                    "about": _(
                        "MSA value for the MSA type specified in the "
                        "'msa_type' column, when the impact value is within "
                        "the range given in the 'value' column.")}
            },
            "about": _(
                "Table that sets the MSA values for each impact driver and "
                "each range of impact values."),
            "name": _("MSA Parameter Table")
        },
        "aoi_path": {
            **spec_utils.AOI,
            "projected": True,
            "required": False
        },
        "globio_lulc_path": {
            **spec_utils.LULC,
            "projected": True,
            "required": "predefined_globio",
            "about": _(
                "Predefined GLOBIO LULC map using the standard GLOBIO "
                "classification scheme and codes. "
                "Required if Use Predefined GLOBIO LULC is selected."),
            "name": _("GLOBIO Classified Land Use")
        }
    }
}


def execute(args):
    """GLOBIO.

    The model operates in two modes.  Mode (a) generates a landcover map
    based on a base landcover map and information about crop yields,
    infrastructure, and more.  Mode (b) assumes the globio landcover
    map is generated.  These modes are used below to describe input
    parameters.

    Args:

        args['workspace_dir'] (string): output directory for intermediate,
            temporary, and final files
        args['predefined_globio'] (boolean): if True then "mode (b)" else
            "mode (a)"
        args['results_suffix'] (string): (optional) string to append to any
            output files
        args['lulc_path'] (string): used in "mode (a)" path to a base landcover
            map with integer codes
        args['lulc_to_globio_table_path'] (string): used in "mode (a)" path to
            table that translates the land-cover args['lulc_path'] to
            intermediate GLOBIO classes, from which they will be further
            differentiated using the additional data in the model.  Contains
            at least the following fields:

            * 'lucode': Land use and land cover class code of the dataset
              used. LULC codes match the 'values' column in the LULC
              raster of mode (b) and must be numeric and unique.
            * 'globio_lucode': The LULC code corresponding to the GLOBIO class
              to which it should be converted, using intermediate codes
              described in the example below.

        args['infrastructure_dir'] (string): used in "mode (a) and (b)" a path
            to a folder containing maps of either gdal compatible rasters or
            OGR compatible shapefiles.  These data will be used in the
            infrastructure to calculation of MSA.
        args['pasture_path'] (string): used in "mode (a)" path to pasture
            raster
        args['potential_vegetation_path'] (string): used in "mode (a)" path to
            potential vegetation raster
        args['pasture_threshold'] (float): used in "mode (a)"
        args['intensification_fraction'] (float): a value between 0 and 1
            denoting proportion of total agriculture that should be classified
            as 'high input'
        args['primary_threshold'] (float): used in "mode (a)"
        args['msa_parameters_path'] (string): path to MSA classification
            parameters
        args['aoi_path'] (string): (optional) if it exists then final MSA
            raster is summarized by AOI
        args['globio_lulc_path'] (string): used in "mode (b)" path to
            predefined globio raster.
        args['n_workers'] (int): (optional) The number of worker processes to
            use for processing this model.  If omitted, computation will take
            place in the current process.

    Returns:
        None

    """
    msa_parameter_table = load_msa_parameter_table(
        args['msa_parameters_path'], float(args['intensification_fraction']))
    file_suffix = utils.make_suffix_string(args, 'results_suffix')
    output_dir = os.path.join(args['workspace_dir'])
    # For intermediate files that users may want to explore:
    intermediate_dir = os.path.join(
        args['workspace_dir'], 'intermediate_outputs')
    # For intermediate files that users probably don't need to see,
    # but should persist for taskgraph purposes:
    tmp_dir = os.path.join(intermediate_dir, 'tmp')
    utils.make_directories(
        [output_dir, intermediate_dir, tmp_dir])

    # Initialize a TaskGraph
    taskgraph_db_dir = os.path.join(intermediate_dir, '_taskgraph_working_dir')
    try:
        n_workers = int(args['n_workers'])
    except (KeyError, ValueError, TypeError):
        # KeyError when n_workers is not present in args
        # ValueError when n_workers is an empty string.
        # TypeError when n_workers is None.
        n_workers = -1  # single process mode.
    task_graph = taskgraph.TaskGraph(taskgraph_db_dir, n_workers)

    gaussian_kernel_path = os.path.join(
        tmp_dir, 'gaussian_kernel%s.tif' % file_suffix)
    make_gaussian_kernel_task = task_graph.add_task(
        func=make_gaussian_kernel_path,
        args=(SIGMA, gaussian_kernel_path),
        target_path_list=[gaussian_kernel_path],
        task_name='gaussian_kernel')

    calculate_globio_task_list = []
    # get base raster cell size and nodata from whichever lulc is
    # provided in args
    if not args['predefined_globio']:
        globio_lulc_path = os.path.join(
            intermediate_dir, 'globio_lulc%s.tif' % file_suffix)
        base_lulc_info = pygeoprocessing.get_raster_info(args['lulc_path'])
        out_pixel_size = (abs(base_lulc_info['pixel_size'][0]) +
                          abs(base_lulc_info['pixel_size'][0])) / 2
        globio_nodata = -1
        globio_lulc_task = _calculate_globio_lulc_map(
            args['lulc_to_globio_table_path'], args['lulc_path'],
            args['potential_vegetation_path'], args['pasture_path'],
            gaussian_kernel_path, float(args['pasture_threshold']),
            float(args['primary_threshold']), file_suffix,
            tmp_dir, globio_lulc_path, globio_nodata, task_graph)
        calculate_globio_task_list.append(globio_lulc_task)
    else:
        LOGGER.info('no need to calculate GLOBIO LULC because it is passed in')
        globio_lulc_path = args['globio_lulc_path']
        globio_lulc_info = pygeoprocessing.get_raster_info(globio_lulc_path)
        out_pixel_size = (abs(globio_lulc_info['pixel_size'][0]) +
                          abs(globio_lulc_info['pixel_size'][0])) / 2
        globio_nodata = globio_lulc_info['nodata'][0]

    infrastructure_path = os.path.join(
        tmp_dir, 'combined_infrastructure%s.tif' % file_suffix)
    combine_infrastructure_task = task_graph.add_task(
        func=_collapse_infrastructure_layers,
        args=(args['infrastructure_dir'], globio_lulc_path,
              infrastructure_path, tmp_dir),
        target_path_list=[infrastructure_path],
        dependent_task_list=calculate_globio_task_list,
        task_name='combine_infrastructure')

    # calc_msa_f
    primary_veg_mask_path = os.path.join(
        tmp_dir, 'primary_veg_mask%s.tif' % file_suffix)
    primary_veg_mask_nodata = -1

    LOGGER.info("create mask of primary veg areas")
    # lucodes for primary veg are hardcoded in the local_op
    mask_primary_veg_task = task_graph.add_task(
        func=pygeoprocessing.raster_calculator,
        args=([(globio_lulc_path, 1), (globio_nodata, 'raw'),
               (primary_veg_mask_nodata, 'raw')], _primary_veg_mask_op,
              primary_veg_mask_path, gdal.GDT_Int16, primary_veg_mask_nodata),
        target_path_list=[primary_veg_mask_path],
        dependent_task_list=calculate_globio_task_list,
        task_name='mask_primary_veg')

    LOGGER.info('smooth primary veg areas with gaussian filter')
    smoothed_primary_veg_mask_path = os.path.join(
        tmp_dir, 'smoothed_primary_veg_mask%s.tif' % file_suffix)
    smooth_primary_veg_mask_task = task_graph.add_task(
        func=pygeoprocessing.convolve_2d,
        args=((primary_veg_mask_path, 1), (gaussian_kernel_path, 1),
              smoothed_primary_veg_mask_path),
        target_path_list=[smoothed_primary_veg_mask_path],
        dependent_task_list=[mask_primary_veg_task, make_gaussian_kernel_task],
        task_name='smooth_primary_veg_mask')

    LOGGER.info('calculate primary_veg_smooth')
    # Passing the filter over the veg mask means veg has bled outside the mask,
    # so mask it again to get the final ffqi
    primary_veg_smooth_path = os.path.join(
        intermediate_dir, 'primary_veg_smooth%s.tif' % file_suffix)
    smooth_primary_veg_task = task_graph.add_task(
        func=pygeoprocessing.raster_calculator,
        args=([(primary_veg_mask_path, 1), (smoothed_primary_veg_mask_path, 1),
               (primary_veg_mask_nodata, 'raw')],
              _ffqi_op, primary_veg_smooth_path, gdal.GDT_Float32,
              primary_veg_mask_nodata),
        target_path_list=[primary_veg_smooth_path],
        dependent_task_list=[smooth_primary_veg_mask_task],
        task_name='smooth_primary_veg')

    LOGGER.info('calculate msa_f')
    msa_nodata = -1
    msa_f_table = msa_parameter_table['msa_f']
    msa_f_path = os.path.join(output_dir, 'msa_f%s.tif' % file_suffix)

    calculate_msa_f_task = task_graph.add_task(
        func=_msa_f_calculation,
        args=(primary_veg_smooth_path, msa_f_table, msa_f_path, msa_nodata),
        target_path_list=[msa_f_path],
        dependent_task_list=[smooth_primary_veg_task],
        task_name='calculate_msa_f')

    # calc_msa_i
    msa_i_other_table = msa_parameter_table['msa_i_other']
    msa_i_primary_table = msa_parameter_table['msa_i_primary']

    LOGGER.info('distance transform infrasture raster')
    distance_to_infrastructure_path = os.path.join(
        intermediate_dir, 'distance_to_infrastructure%s.tif' % file_suffix)
    distance_to_infrastructure_task = task_graph.add_task(
        func=pygeoprocessing.distance_transform_edt,
        args=((infrastructure_path, 1), distance_to_infrastructure_path),
        target_path_list=[distance_to_infrastructure_path],
        dependent_task_list=[combine_infrastructure_task],
        task_name='distance_to_infrastructure')

    LOGGER.info('calculate msa_i')
    msa_i_path = os.path.join(output_dir, 'msa_i%s.tif' % file_suffix)
    calculate_msa_i_task = task_graph.add_task(
        func=_msa_i_calculation,
        args=(globio_lulc_path, distance_to_infrastructure_path,
              out_pixel_size, msa_i_primary_table, msa_i_other_table,
              msa_i_path, msa_nodata),
        target_path_list=[msa_i_path],
        dependent_task_list=[distance_to_infrastructure_task],
        task_name='calculate_msa_i')

    # calc_msa_lu
    msa_lu_path = os.path.join(
        output_dir, 'msa_lu%s.tif' % file_suffix)
    LOGGER.info('calculate msa_lu')
    reclass_error_details = {
        'raster_name': 'GLOBIO LULC', 'column_name': 'MSA_type-msa_lu',
        'table_name': 'MSA'}
    calculate_msa_lu_task = task_graph.add_task(
        func=utils.reclassify_raster,
        args=((globio_lulc_path, 1), msa_parameter_table['msa_lu'],
              msa_lu_path, gdal.GDT_Float32, globio_nodata,
              reclass_error_details),
        target_path_list=[msa_lu_path],
        dependent_task_list=calculate_globio_task_list,
        task_name='calculate_msa_lu')

    LOGGER.info('calculate msa')
    msa_path = os.path.join(
        output_dir, 'msa%s.tif' % file_suffix)
    calculate_msa_task = task_graph.add_task(
        func=_msa_calculation,
        args=(msa_f_path, msa_lu_path, msa_i_path, msa_path, msa_nodata),
        target_path_list=[msa_path],
        dependent_task_list=[
            calculate_msa_f_task, calculate_msa_i_task, calculate_msa_lu_task],
        task_name='calculate_msa')

    LOGGER.info('summarize msa result in AOI polygons')
    # the AOI is an optional argument, so check for its existence
    if 'aoi_path' in args and len(args['aoi_path']) > 0:
        summary_aoi_path = os.path.join(
            output_dir, 'aoi_summary%s.shp' % file_suffix)
        task_graph.add_task(
            func=_summarize_results_in_aoi,
            args=(args['aoi_path'], summary_aoi_path, msa_path),
            target_path_list=[summary_aoi_path],
            dependent_task_list=[calculate_msa_task],
            task_name='summarize_msa_in_aoi')

    task_graph.close()
    task_graph.join()


def _summarize_results_in_aoi(aoi_path, summary_aoi_path, msa_path):
    """Aggregate MSA results to AOI polygons with zonal statistics.

    Args:
        aoi_path (string): path to aoi shapefile containing polygons.
        summary_aoi_path (string):
            path to copy of aoi shapefile with summary stats added.
        msa_path (string): path to msa results raster to summarize.

    Returns:
        None

    """
    # copy the aoi to an output shapefile
    original_datasource = gdal.OpenEx(
        aoi_path, gdal.OF_VECTOR | gdal.GA_ReadOnly)
    # Delete if existing shapefile with the same name
    if os.path.isfile(summary_aoi_path):
        os.remove(summary_aoi_path)
    # Copy the input shapefile into the designated output folder
    driver = gdal.GetDriverByName('ESRI Shapefile')
    datasource_copy = driver.CreateCopy(
        summary_aoi_path, original_datasource)
    layer = datasource_copy.GetLayer()
    msa_summary_field_def = ogr.FieldDefn('msa_mean', ogr.OFTReal)
    msa_summary_field_def.SetWidth(24)
    msa_summary_field_def.SetPrecision(11)
    layer.CreateField(msa_summary_field_def)
    layer.SyncToDisk()

    msa_summary = pygeoprocessing.zonal_statistics(
        (msa_path, 1), summary_aoi_path)
    for feature in layer:
        feature_fid = feature.GetFID()
        # count == 0 if polygon outside raster bounds or only over nodata
        if msa_summary[feature_fid]['count'] != 0:
            field_val = (
                float(msa_summary[feature_fid]['sum'])
                / float(msa_summary[feature_fid]['count']))
            feature.SetField('msa_mean', field_val)
            layer.SetFeature(feature)


def _primary_veg_mask_op(lulc_array, globio_nodata, primary_veg_mask_nodata):
    """Masking out natural areas."""
    # lulc_array and nodata could conceivably be a float here,
    # if it's the user-provided globio dataset
    # landcover type 1 in the GLOBIO schema represents primary vegetation
    result = numpy.empty_like(lulc_array, dtype=numpy.int16)
    result[:] = primary_veg_mask_nodata
    valid_mask = slice(None)
    if globio_nodata is not None:
        valid_mask = ~utils.array_equals_nodata(lulc_array, globio_nodata)
    result[valid_mask] = lulc_array[valid_mask] == 1
    return result


def _ffqi_op(forest_areas_array, smoothed_forest_areas, forest_areas_nodata):
    """Mask out ffqi only where there's an ffqi."""
    result = numpy.empty_like(forest_areas_array, dtype=numpy.float32)
    result[:] = forest_areas_nodata
    # forest_areas_array and _nodata are integer types and not user-defined
    valid_mask = ~utils.array_equals_nodata(
        forest_areas_array, forest_areas_nodata)
    result[valid_mask] = (
        forest_areas_array[valid_mask] * smoothed_forest_areas[valid_mask])
    return result


def _msa_f_calculation(
        primary_veg_smooth_path, msa_f_table, msa_f_path, msa_nodata):
    """Calculate msa fragmentation.

    Bin ffqi values based on rules defined in msa_parameters.csv.

    Args:
        primary_veg_smooth (str): path to a raster with float values
            representing ffqi.
        msa_f_table (dict):
            subset of msa_parameters.csv with fragmentation bins defined.
        msa_nodata (int/float): output nodata value

    Returns:
        Nothing
    """
    primary_veg_info = pygeoprocessing.get_raster_info(primary_veg_smooth_path)
    primary_veg_mask_nodata = primary_veg_info['nodata'][0]

    msa_f_table_copy = msa_f_table.copy()
    less_than = msa_f_table_copy.pop('<', None)
    greater_than = msa_f_table_copy.pop('>', None)

    def msa_f_op(primary_veg_smooth):
        """Calculate msa fragmentation.

        Bin ffqi values based on rules defined in msa_parameters.csv.

        Args:
            primary_veg_smooth (array): float values representing ffqi.

        Returns:
            Array with float values. One component of final MSA score.
        """
        msa_f = numpy.full_like(
            primary_veg_smooth, msa_nodata, dtype=numpy.float32)

        if greater_than:
            msa_f[primary_veg_smooth > greater_than[0]] = (
                greater_than[1])
        for key in reversed(sorted(msa_f_table_copy)):
            msa_f[primary_veg_smooth <= key] = msa_f_table_copy[key]
        if less_than:
            msa_f[primary_veg_smooth < less_than[0]] = (
                less_than[1])

        if msa_nodata is not None:
            nodata_mask = utils.array_equals_nodata(
                primary_veg_smooth, primary_veg_mask_nodata)
            msa_f[nodata_mask] = msa_nodata

        return msa_f

    pygeoprocessing.raster_calculator(
        [(primary_veg_smooth_path, 1)], msa_f_op, msa_f_path, gdal.GDT_Float32,
        msa_nodata)


def _msa_i_calculation(
        globio_lulc_path, distance_to_infrastructure_path, out_pixel_size,
        msa_i_primary_table, msa_i_other_table, msa_i_path, msa_nodata):
    """Calculate msa infrastructure.

    Bin distance_to_infrastructure values according to rules defined
    in msa_parameters.csv.

    Args:
        globio_lulc_path (str): path to raster with globio landcover codes.
        distance_to_infrastructure_path (str): path to raster with
            float values measuring distance from nearest infrastructure present
            in layers from args['infrastructure_dir'].
        out_pixel_size (float): from the globio lulc raster info.
        msa_i_primary_table (dict):
            subset of msa_parameters.csv with distance to infrastructure bins
            defined. These bins are applied to areas of primary veg.
        msa_i_other_table (dict):
            subset of msa_parameters.csv with distance to infrastructure bins
            defined. These bins are applied to areas of not primary veg.
        msa_i_path (str): output path for msa infrastructure raster.
        msa_nodata (float): output nodata value.

    Returns:
        Nothing.
    """
    lulc_info = pygeoprocessing.get_raster_info(globio_lulc_path)
    lulc_nodata = lulc_info['nodata'][0]

    # Create a copy of the dictionary so we don't mutate it outside this scope
    msa_i_primary_table_copy = msa_i_primary_table.copy()
    msa_i_other_table_copy = msa_i_other_table.copy()

    primary_less_than = msa_i_primary_table_copy.pop('<', None)
    primary_greater_than = msa_i_primary_table_copy.pop('>', None)
    other_less_than = msa_i_other_table_copy.pop('<', None)
    other_greater_than = msa_i_other_table_copy.pop('>', None)

    def msa_i_op(lulc_array, distance_to_infrastructure):
        """Calculate msa infrastructure.

        Args:
            lulc_array (array): integer values representing globio landcover
                codes.
            distance_to_infrastructure (array): float values measuring
                distance from nearest infrastructure present in layers from
                args['infrastructure_dir'].

        Returns:
            Array with float values. One component of final MSA score.
        """
        distance_to_infrastructure *= out_pixel_size  # convert to meters
        # Use `full_like` with `msa_nodata` value because we can't know if
        # entire range of values will be covered from msa tables
        msa_i_primary = numpy.full_like(
            lulc_array, msa_nodata, dtype=numpy.float32)
        msa_i_other = numpy.full_like(
            lulc_array, msa_nodata, dtype=numpy.float32)
        nodata_mask = utils.array_equals_nodata(lulc_array, lulc_nodata)

        if primary_greater_than:
            msa_i_primary[distance_to_infrastructure > primary_greater_than[0]] = (
                primary_greater_than[1])
        for key in reversed(sorted(msa_i_primary_table_copy)):
            msa_i_primary[distance_to_infrastructure <= key] = (
                msa_i_primary_table_copy[key])
        if primary_less_than:
            msa_i_primary[distance_to_infrastructure < primary_less_than[0]] = (
                primary_less_than[1])

        if other_greater_than:
            msa_i_other[distance_to_infrastructure > other_greater_than[0]] = (
                other_greater_than[1])
        for key in reversed(sorted(msa_i_other_table_copy)):
            msa_i_other[distance_to_infrastructure <= key] = (
                msa_i_other_table_copy[key])
        if other_less_than:
            msa_i_other[distance_to_infrastructure < other_less_than[0]] = (
                other_less_than[1])

        # lulc code 1 is primary veg
        msa_i = numpy.where(lulc_array == 1, msa_i_primary, msa_i_other)
        msa_i[nodata_mask] = msa_nodata
        return msa_i

    pygeoprocessing.raster_calculator(
        [(globio_lulc_path, 1), (distance_to_infrastructure_path, 1)],
        msa_i_op, msa_i_path, gdal.GDT_Float32, msa_nodata)


def _msa_calculation(
        msa_f_path, msa_lu_path, msa_i_path, msa_path, msa_nodata):
    """Calculate the MSA which is the product of the sub MSAs.

    Args:
        msa_f_path (str): path to the msa_f raster.
        msa_lu_path (str): path to the msa_lu raster.
        msa_i_path (str): path to the msa_i raster.
        msa_path (str): path to the output MSA raster.
        msa_nodata (int/float): the output nodata value.

    Returns:
        Nothing
    """
    msa_f_nodata = pygeoprocessing.get_raster_info(msa_f_path)['nodata'][0]
    msa_lu_nodata = pygeoprocessing.get_raster_info(msa_lu_path)['nodata'][0]
    msa_i_nodata = pygeoprocessing.get_raster_info(msa_i_path)['nodata'][0]
    nodata_array = [msa_f_nodata, msa_lu_nodata, msa_i_nodata]

    def msa_op(msa_f, msa_lu, msa_i):
        """Calculate the MSA which is the product of the sub MSAs."""
        result = numpy.full_like(msa_f, msa_nodata, dtype=numpy.float32)
        valid_mask = numpy.ones(msa_f.shape, dtype=bool)
        for msa_array, nodata_val in zip([msa_f, msa_lu, msa_i], nodata_array):
            if nodata_val is not None:
                valid_mask &= ~utils.array_equals_nodata(msa_array, nodata_val)
        result[valid_mask] = (
            msa_f[valid_mask] * msa_lu[valid_mask] * msa_i[valid_mask])
        return result

    pygeoprocessing.raster_calculator(
        [(msa_f_path, 1), (msa_lu_path, 1), (msa_i_path, 1)],
        msa_op, msa_path, gdal.GDT_Float32, msa_nodata)


def make_gaussian_kernel_path(sigma, kernel_path):
    """Create a gaussian kernel raster."""
    max_distance = sigma * 5
    kernel_size = int(numpy.round(max_distance * 2 + 1))

    driver = gdal.GetDriverByName('GTiff')
    kernel_dataset = driver.Create(
        kernel_path.encode('utf-8'), kernel_size, kernel_size, 1,
        gdal.GDT_Float32, options=['BIGTIFF=IF_SAFER'])

    # Make some kind of geotransform, it doesn't matter what but
    # will make GIS libraries behave better if it's all defined
    kernel_dataset.SetGeoTransform([444720, 30, 0, 3751320, 0, -30])
    srs = osr.SpatialReference()
    srs.SetUTM(11, 1)
    srs.SetWellKnownGeogCS('NAD27')
    kernel_dataset.SetProjection(srs.ExportToWkt())

    kernel_band = kernel_dataset.GetRasterBand(1)
    kernel_band.SetNoDataValue(-9999)

    col_index = numpy.array(range(kernel_size))
    integration = 0.0
    for row_index in range(kernel_size):
        kernel = numpy.exp(
            -((row_index - max_distance)**2 +
              (col_index - max_distance) ** 2)/(2.0*sigma**2)).reshape(
                  1, kernel_size)

        integration += numpy.sum(kernel)
        kernel_band.WriteArray(kernel, xoff=0, yoff=row_index)

    for row_index in range(kernel_size):
        kernel_row = kernel_band.ReadAsArray(
            xoff=0, yoff=row_index, win_xsize=kernel_size, win_ysize=1)
        kernel_row /= integration
        kernel_band.WriteArray(kernel_row, 0, row_index)


def load_msa_parameter_table(
        msa_parameter_table_filename, intensification_fraction):
    """Load parameter table to a dict that to define the MSA ranges.

    Args:
        msa_parameter_table_filename (string): path to msa csv table
        intensification_fraction (float): a number between 0 and 1 indicating
            what level between msa_lu 8 and 9 to define the general GLOBIO
            code "12" to.

    Returns:
        dictionary of the form:

            {'msa_f': {
                valuea: msa_f_value, ...
                valueb: ...
                '<': (bound, msa_f_value),
                '>': (bound, msa_f_value)}
            'msa_i_other_table': {
                valuea: msa_i_value, ...
                valueb: ...
                '<': (bound, msa_i_other_value),
                '>': (bound, msa_i_other_value)}
            'msa_i_primary': {
                valuea: msa_i_primary_value, ...
                valueb: ...
                '<': (bound, msa_i_primary_value),
                '>': (bound, msa_i_primary_value)}
            'msa_lu': {
                valuea: msa_lu_value, ...
                valueb: ...
                '<': (bound, msa_lu_value),
                '>': (bound, msa_lu_value)
                12: (msa_lu_8 * (1.0 - intensification_fraction) +
                msa_lu_9 * intensification_fraction}}

    """
    msa_table = utils.read_csv_to_dataframe(
        msa_parameter_table_filename, sep=None)
    msa_dict = collections.defaultdict(dict)
    for _, row in msa_table.iterrows():
        if row['Value'][0] in ['<', '>']:
            # put the limit and the MSA value in a tub
            value = row['Value'][0]
            # take 1: because it starts with a < or >
            msa_dict[row['MSA_type']][value] = (
                float(row['Value'][1:]), float(row['MSA_x']))
            continue
        elif '-' in row['Value']:
            value = float(row['Value'].split('-')[1])
        else:
            value = float(row['Value'])
        msa_dict[row['MSA_type']][value] = float(row['MSA_x'])
    # landcover ID 12 is a linear interpolation between 8 and 9
    msa_dict['msa_lu'][12] = (
        msa_dict['msa_lu'][8] * (1.0 - intensification_fraction) +
        msa_dict['msa_lu'][9] * intensification_fraction)
    return dict(msa_dict)


def _calculate_globio_lulc_map(
        lulc_to_globio_table_path, lulc_path, potential_vegetation_path,
        pasture_path, gaussian_kernel_path, pasture_threshold,
        primary_threshold, file_suffix, tmp_dir,
        globio_lulc_path, globio_nodata, task_graph):
    """Translate a general landcover map into a GLOBIO version.

    Args:
        lulc_to_globio_table_path (string): a table that maps arbitrary
            landcover values to globio equivalents.
        lulc_path (string): path to the raw landcover map.
        potential_vegetation_path (string): a landcover map that indicates what
            the vegetation types would be if left to revert to natural state
        pasture_path (string): a path to a raster that indicates the percent
            of pasture contained in the pixel.  used to classify forest types
            from scrubland.
        gaussian_kernel_path (string): path to gaussian kernel raster
            passed to convolution.
        pasture_threshold (float): the threshold to classify pixels in pasture
            as potential forest or scrub
        primary_threshold (float): the threshold to classify the calculated
            FFQI pixels into core forest or secondary
        file_suffix - (string) to append on output file
        tmp_dir (string): path to location for intermediate files that users
            probably don't need to see, but should persist for taskgraph.
            The following files are created:
                'intermediate_globio_lulc.tif': reclassified landcover map
                    to globio landcover codes
                'ffqi.tif': index of fragmentation due to infrastructure and
                    original values of landscape
        globio_lulc_path (string): Path to globio lulc raster. Primary output
            of the function, starts with intermeidate globio and modifies based
            on the other biophysical parameters to the function as described in
            the GLOBIO process.
        globio_nodata (int): nodata value assigned to globio_lulc_path raster
        task_graph (TaskGraph): in-memory object from taskgraph.TaskGraph()

    Returns:
        The ultimate task in this branch of the task_graph

    """
    lulc_to_globio_table = utils.build_lookup_from_csv(
        lulc_to_globio_table_path, 'lucode')

    lulc_to_globio = dict(
        [(lulc_code, int(table['globio_lucode'])) for
         (lulc_code, table) in lulc_to_globio_table.items()])

    intermediate_globio_lulc_path = os.path.join(
        tmp_dir, 'intermediate_globio_lulc%s.tif' % file_suffix)
    reclass_error_details = {
        'raster_name': 'LULC', 'column_name': 'lucode',
        'table_name': 'Land Cover to GLOBIO Land Cover'}
    reclass_lulc_to_globio_task = task_graph.add_task(
        func=utils.reclassify_raster,
        args=((lulc_path, 1), lulc_to_globio, intermediate_globio_lulc_path,
              gdal.GDT_Int32, globio_nodata, reclass_error_details),
        target_path_list=[intermediate_globio_lulc_path],
        task_name='reclassify_lulc_to_globio')

    forest_areas_path = os.path.join(
        tmp_dir, 'forest_areas%s.tif' % file_suffix)
    forest_areas_nodata = -1

    LOGGER.info("create mask of natural areas")
    mask_forests_task = task_graph.add_task(
        func=pygeoprocessing.raster_calculator,
        args=([(intermediate_globio_lulc_path, 1), (globio_nodata, 'raw'),
               (forest_areas_nodata, 'raw')], _forest_area_mask_op,
              forest_areas_path, gdal.GDT_Int32, forest_areas_nodata),
        target_path_list=[forest_areas_path],
        dependent_task_list=[reclass_lulc_to_globio_task],
        task_name='mask_forest_area')

    LOGGER.info('smooth natural areas with gaussian filter')
    smoothed_forest_areas_path = os.path.join(
        tmp_dir, 'smoothed_forest_areas%s.tif' % file_suffix)
    smooth_forest_areas_task = task_graph.add_task(
        func=pygeoprocessing.convolve_2d,
        args=((forest_areas_path, 1), (gaussian_kernel_path, 1),
              smoothed_forest_areas_path),
        target_path_list=[smoothed_forest_areas_path],
        dependent_task_list=[mask_forests_task],
        task_name='smooth_forest_areas')

    ffqi_path = os.path.join(
        tmp_dir, 'ffqi%s.tif' % file_suffix)
    LOGGER.info('calculate ffqi')
    calculate_ffqi_task = task_graph.add_task(
        func=pygeoprocessing.raster_calculator,
        args=([(forest_areas_path, 1), (smoothed_forest_areas_path, 1),
               (forest_areas_nodata, 'raw')], _ffqi_op,
              ffqi_path, gdal.GDT_Float32, forest_areas_nodata),
        target_path_list=[ffqi_path],
        dependent_task_list=[smooth_forest_areas_task],
        task_name='ffqi')

    LOGGER.info('create the globio lulc')
    # The veg and pasture rasters are user inputs,
    # so may not be aligned yet with lulc
    base_raster_align_list = [potential_vegetation_path, pasture_path]
    target_raster_align_list = [os.path.join(
        tmp_dir, os.path.basename(x))
        for x in base_raster_align_list]
    base_raster_info = pygeoprocessing.get_raster_info(lulc_path)
    align_veg_pasture_task = task_graph.add_task(
        func=pygeoprocessing.align_and_resize_raster_stack,
        args=(base_raster_align_list,
              target_raster_align_list,
              ['near', 'bilinear'],
              base_raster_info['pixel_size'],
              base_raster_info['bounding_box']),
        target_path_list=target_raster_align_list,
        task_name='align_veg_pasture_rasters')

    calculate_globio_lulc_task = task_graph.add_task(
        func=pygeoprocessing.raster_calculator,
        args=([(intermediate_globio_lulc_path, 1),
               (target_raster_align_list[0], 1),
               (target_raster_align_list[1], 1),
               (ffqi_path, 1), (globio_nodata, 'raw'),
               (pasture_threshold, 'raw'),
               (primary_threshold, 'raw')],
              _create_globio_lulc_op, globio_lulc_path, gdal.GDT_Int16,
              globio_nodata),
        target_path_list=[globio_lulc_path],
        dependent_task_list=[align_veg_pasture_task, calculate_ffqi_task],
        task_name='calculate_globio_lulc')

    return calculate_globio_lulc_task


def _forest_area_mask_op(lulc_array, globio_nodata, forest_areas_nodata):
    """Masking out forest areas."""
    # comparing integers, numpy.isclose not needed
    valid_mask = ~utils.array_equals_nodata(lulc_array, globio_nodata)
    # landcover code 130 represents all MODIS forest codes which originate
    # as 1-5
    result = numpy.empty_like(lulc_array, dtype=numpy.int16)
    result[:] = forest_areas_nodata
    result[valid_mask] = lulc_array[valid_mask] == 130
    return result


def _create_globio_lulc_op(
        lulc_array, potential_vegetation_array, pasture_array,
        ffqi, globio_nodata, pasture_threshold, primary_threshold):
    """Construct GLOBIO lulc given relevant biophysical parameters."""
    result = numpy.empty_like(lulc_array, dtype=numpy.int16)
    result[:] = globio_nodata
    valid_mask = ~utils.array_equals_nodata(lulc_array, globio_nodata)
    valid_result = result[valid_mask]

    # Split Shrublands and grasslands into primary vegetations,
    # livestock grazing areas, and man-made pastures. Landcover
    # 131 represents grassland/shrubland in the GLOBIO classification.
    grass_shrub_mask = lulc_array[valid_mask] == 131
    grass_shrub_result = valid_result[grass_shrub_mask]
    # fill with livestock grazing, then re-assign to pasture, primary veg.
    grass_shrub_result[:] = 5

    # man-made pasture
    valid_pasture_mask = (
        potential_vegetation_array[valid_mask][grass_shrub_mask] <= 8)
    grass_shrub_result[valid_pasture_mask] = 6

    # primary vegetation
    valid_primary_veg_mask = ~valid_pasture_mask & (
        pasture_array[valid_mask][grass_shrub_mask] < pasture_threshold)
    grass_shrub_result[valid_primary_veg_mask] = 1

    valid_result[grass_shrub_mask] = grass_shrub_result

    # Outside of the grass/shrub categories, carry over the original codes:
    valid_result[~grass_shrub_mask] = lulc_array[valid_mask][~grass_shrub_mask]

    # Step 1.4a: Split Forests into Primary, Secondary
    # 1 is primary forest
    # 3 is secondary forest
    valid_modis_forest_mask = lulc_array[valid_mask] == 130
    forest_result = valid_result[valid_modis_forest_mask]
    forest_result[:] = 1
    forest_result[
        ffqi[valid_mask][valid_modis_forest_mask] < primary_threshold] = 3
    valid_result[valid_modis_forest_mask] = forest_result

    # Classify all ag classes as a new LULC value "12" per our custom design
    # of agriculture. Landcover 132 represents agriculture landcover types
    # in the GLOBIO classification scheme
    valid_ag_mask = lulc_array[valid_mask] == 132
    valid_result[valid_ag_mask] = 12

    result[valid_mask] = valid_result
    return result


def _collapse_infrastructure_layers(
        infrastructure_dir, base_raster_path, infrastructure_path,
        tmp_dir):
    """Collapse all GIS infrastructure layers to one raster.

    Gathers all the GIS layers in the given directory and collapses them
    to a single byte raster mask where 1 indicates a pixel overlapping with
    one of the original infrastructure layers, 0 does not, and nodata
    indicates a region that has no layers that overlap but are still contained
    in the bounding box.

    Args:
        infrastructure_dir (string): path to a directory containing maps of
            either gdal compatible rasters or OGR compatible shapefiles.
        base_raster_path (string): a path to a file that has the dimensions and
            projection of the desired output infrastructure file.
        infrastructure_path (string): (output) path to a file that will be a
            byte raster with 1s everywhere there was a GIS layer present in
            the GIS layers in `infrastructure_dir`.
        tmp_dir (string): path to folder to store inetermediate datasets such
            as aligned versions of infrastructure rasters.

    Returns:
        None

    """
    # load the infrastructure layers from disk
    infrastructure_filenames = []
    infrastructure_nodata_list = []
    infrastructure_tmp_filenames = []
    # in case we need to rasterize some vector inputs:
    tmp_rasterize_dir = os.path.join(tmp_dir, 'rasterized')
    for root_directory, _, filename_list in os.walk(infrastructure_dir):
        for filename in filename_list:
            if filename.lower().endswith(".tif"):
                infrastructure_filenames.append(
                    os.path.join(root_directory, filename))
                infrastructure_nodata_list.append(
                    pygeoprocessing.get_raster_info(
                        infrastructure_filenames[-1])['nodata'][0])

            if filename.lower().endswith(".shp"):
                utils.make_directories([tmp_rasterize_dir])
                file_handle, tmp_raster_path = tempfile.mkstemp(
                    dir=tmp_rasterize_dir, suffix='.tif')
                os.close(file_handle)

                pygeoprocessing.new_raster_from_base(
                    base_raster_path, tmp_raster_path,
                    gdal.GDT_Int32, [-1.0], fill_value_list=[0])
                pygeoprocessing.rasterize(
                    os.path.join(root_directory, filename),
                    tmp_raster_path, burn_values=[1],
                    option_list=["ALL_TOUCHED=TRUE"])

                infrastructure_filenames.append(tmp_raster_path)
                infrastructure_tmp_filenames.append(tmp_raster_path)
                infrastructure_nodata_list.append(
                    pygeoprocessing.get_raster_info(
                        infrastructure_filenames[-1])['nodata'][0])

    if len(infrastructure_filenames) == 0:
        raise ValueError(
            "infrastructure directory didn't have any rasters or "
            "vectors at %s", infrastructure_dir)

    infrastructure_nodata = 255

    def _collapse_infrastructure_op(*infrastructure_array_list):
        """For each pixel, create mask 1 if all valid, else set to nodata."""
        # Zeroes indicate where there's no infrastructure influence but not
        # necessarily nodata
        infrastructure_result = numpy.zeros(
            infrastructure_array_list[0].shape, dtype=numpy.uint8)

        nodata_mask = numpy.full(infrastructure_array_list[0].shape, True)
        infrastructure_mask = numpy.full(
            infrastructure_array_list[0].shape, False)

        for index in range(len(infrastructure_array_list)):
            # mark all pixels True that have infrastructure in this layer
            infrastructure_mask |= infrastructure_array_list[index] > 0
            if infrastructure_nodata_list[index] is not None:
                # update nodata mask: intersection with this layer
                nodata_mask &= utils.array_equals_nodata(
                    infrastructure_array_list[index],
                    infrastructure_nodata_list[index])
            # if nodata is None, every pixel in this layer has data,
            # so the nodata_mask should be all False
            else:
                nodata_mask &= False

        infrastructure_result[infrastructure_mask] = 1
        infrastructure_result[nodata_mask] = infrastructure_nodata
        return infrastructure_result

    LOGGER.info('collapse infrastructure into one raster')
    aligned_infrastructure_target_list = [os.path.join(
        tmp_dir, os.path.basename(x))
        for x in infrastructure_filenames]
    base_raster_info = pygeoprocessing.get_raster_info(
        base_raster_path)

    pygeoprocessing.align_and_resize_raster_stack(
        infrastructure_filenames,
        aligned_infrastructure_target_list,
        ['near'] * len(infrastructure_filenames),
        base_raster_info['pixel_size'],
        base_raster_info['bounding_box'])
    infra_filename_band_list = [
        (x, 1) for x in aligned_infrastructure_target_list]
    pygeoprocessing.raster_calculator(
        infra_filename_band_list, _collapse_infrastructure_op,
        infrastructure_path, gdal.GDT_Byte, infrastructure_nodata)

    # clean up the temporary filenames
    if os.path.isdir(tmp_rasterize_dir):
        for filename in infrastructure_tmp_filenames:
            os.remove(filename)
        os.rmdir(tmp_rasterize_dir)


@validation.invest_validator
def validate(args, limit_to=None):
    """Validate args to ensure they conform to `execute`'s contract.

    Args:
        args (dict): dictionary of key(str)/value pairs where keys and
            values are specified in `execute` docstring.
        limit_to (str): (optional) if not None indicates that validation
            should only occur on the args[limit_to] value. The intent that
            individual key validation could be significantly less expensive
            than validating the entire `args` dictionary.

    Returns:
        list of ([invalid key_a, invalid_keyb, ...], 'warning/error message')
            tuples. Where an entry indicates that the invalid keys caused
            the error message in the second part of the tuple. This should
            be an empty list if validation succeeds.

    """
    return validation.validate(
        args, ARGS_SPEC['args'], ARGS_SPEC['args_with_spatial_overlap'])



# cython: profile=True
# cython: language_level=2
"""A hierarchical spatial index for fast culling of points in 2D space."""

import os
import sys
import time
import struct
import pickle
import uuid
import itertools
import collections
import traceback
import bisect
import operator
import shapely.geometry
import numpy
import logging

from osgeo import ogr
from osgeo import osr
cimport numpy

MAX_BYTES_TO_BUFFER = 2**27  # buffer a little over 128 megabytes
import buffered_numpy_disk_map
_ARRAY_TUPLE_TYPE = (
    buffered_numpy_disk_map.BufferedNumpyDiskMap._ARRAY_TUPLE_TYPE)

LOGGER = logging.getLogger(
    'natcap.invest.recmodel_server.out_of_core_quadtree')


class OutOfCoreQuadTree(object):
    """An out of core quad tree spatial indexing structure."""

    next_available_blob_id = 0

    def __init__(
            self, bounding_box, max_points_per_node, max_node_depth,
            quad_tree_storage_dir, node_depth=0, node_data_manager=None,
            pickle_filename=None):
        """Make a new quadtree node with a given initial_bounding_box range.

        Args:
            bounding_box (list): list of [x_min, y_min x_max, y_max]
            max_points_per_node (int): maximum number of points before a node
                is subdivided
            max_node_depth (int): max nodes to make the quadtree
            quad_tree_storage_dir (string): path to a directory where the
                quadtree files can be stored
            node_depth (int): depth of current node
            node_data_manager (BufferedNumpyDiskMap): an object which is used
                to store the node data across the entire quadtree
            pickle_filename (string): name of file on disk which to pickle the
                tree to during a flush

        Returns:
            None
        """
        self.bounding_box = list(bounding_box)  # make a copy to avoid aliasing
        self.max_points_per_node = max_points_per_node
        self.max_node_depth = max_node_depth
        self.node_depth = node_depth
        self.n_points_in_node = 0
        self.is_leaf = True  # boolean to determine if this node stores data
        self.nodes = None  # a list of children on an internal node
        self.quad_tree_storage_dir = quad_tree_storage_dir
        if node_data_manager is None:
            self.node_data_manager = (
                buffered_numpy_disk_map.BufferedNumpyDiskMap(
                    pickle_filename+'.db', MAX_BYTES_TO_BUFFER))
        else:
            self.node_data_manager = node_data_manager

        self.pickle_filename = pickle_filename

        # unique blob_id
        self.blob_id = OutOfCoreQuadTree.next_available_blob_id
        OutOfCoreQuadTree.next_available_blob_id += 1

    def flush(self):
        """Flush any cached data to disk."""
        self.node_data_manager.flush()
        if self.pickle_filename is not None:
            pickle.dump(self, open(self.pickle_filename, 'wb'))

    def build_node_shapes(self, ogr_polygon_layer):
        """Add features to an ogr.Layer to visualize quadtree segmentation.

        Args:
            ogr_polygon_layer (ogr.layer): an ogr polygon layer with fields
                'n_points' (int) and 'bb_box' (string) defined.
        Returns:
            None
        """
        # Create a new feature, setting the field and geometry
        if self.is_leaf:
            ring = ogr.Geometry(ogr.wkbLinearRing)
            ring.AddPoint(self.bounding_box[0], self.bounding_box[3])
            ring.AddPoint(self.bounding_box[0], self.bounding_box[1])
            ring.AddPoint(self.bounding_box[2], self.bounding_box[1])
            ring.AddPoint(self.bounding_box[2], self.bounding_box[3])
            ring.AddPoint(self.bounding_box[0], self.bounding_box[3])
            poly = ogr.Geometry(ogr.wkbPolygon)
            poly.AddGeometry(ring)
            feature = ogr.Feature(ogr_polygon_layer.GetLayerDefn())
            feature.SetGeometry(poly)
            feature.SetField('n_points', self.n_points_in_node)
            feature.SetField('bb_box', str(self.bounding_box))
            ogr_polygon_layer.CreateFeature(feature)
        else:
            for node_index in xrange(4):
                self.nodes[node_index].build_node_shapes(ogr_polygon_layer)

    def _get_points_from_node(self):
        """Return points in current node as a list of tuples.

        Returns:
            numpy.ndarray of the form [(data, x_coord, y_coord), ...]
        """
        return self.node_data_manager.read(self.blob_id)

    def _drain_node(self):
        """Delete current node data from the quadtree and return as a list.

        Returns:
            list of (data, x_coord, y_coord)
        """
        userday_tuples = self._get_points_from_node()

        # delete the file because it's drained
        self.node_data_manager.delete(self.blob_id)
        self.blob_id = None
        return userday_tuples

    def _split_node(self):
        """Split node into quads and distribute current node's points."""
        point_list = self._drain_node()
        mid_x_coord = (self.bounding_box[0] + self.bounding_box[2]) / 2.
        mid_y_coord = (self.bounding_box[1] + self.bounding_box[3]) / 2.

        # children node indexes are arranged like this:
        # 01
        # 23
        bounding_quads = [
            [self.bounding_box[0],  # xmin
             mid_y_coord,  # ymin
             mid_x_coord,  # xmax
             self.bounding_box[3]],
            [mid_x_coord,  # xmin
             mid_y_coord,  # ymin
             self.bounding_box[2],  # xmax
             self.bounding_box[3]],
            [self.bounding_box[0],  # xmin
             self.bounding_box[1],  # ymin
             mid_x_coord,  # xmax
             mid_y_coord],  # ymax
            [mid_x_coord,  # xmin
             self.bounding_box[1],  # ymin
             self.bounding_box[2],  # xmax
             mid_y_coord],  # ymax
        ]
        self.nodes = []
        for bounding_box in bounding_quads:
            self.nodes.append(OutOfCoreQuadTree(
                bounding_box,
                self.max_points_per_node,
                self.max_node_depth,
                self.quad_tree_storage_dir,
                node_depth=self.node_depth+1,
                node_data_manager=self.node_data_manager))

        self.n_points_in_node = 0  # points are drained out of this node
        self.is_leaf = False
        # when we add these points to the node they'll be sorted and passed
        # to subquads
        self.add_points(point_list, 0, point_list.size)

    def add_points(
            self, numpy.ndarray point_list, int left_bound, int right_bound):
        """Add a list of points to the current node.

        This function will split the current node if the added points exceed
        the maximum number of points allowed per node and is already not at
        the maximum level.

        Args:
            point_list (numpy.ndarray): a numpy array of
                (data, x_coord, y_coord) tuples
            left_bound (int): left index inclusive of points to consider under
                `point_list`
            right_bound (int): right index non-inclusive of points to consider
                under `point_list`

        Returns:
            None
        """
        cdef float x_coord, y_coord, mid_x_coord, mid_y_coord
        cdef int point_list_len

        if right_bound == left_bound:
            return

        point_list_len = right_bound - left_bound
        if self.is_leaf:
            if ((point_list_len + self.n_points_in_node) <=
                    self.max_points_per_node or
                    self.node_depth == self.max_node_depth):
                # if it's a leaf and the points fit, append to file end
                self.n_points_in_node += point_list_len
                self.node_data_manager.append(
                    self.blob_id, point_list[left_bound:right_bound])
                return  # this lets us pass through the two cases below
            else:
                # this cases handles a leaf node that needs to be split
                self._split_node()

        # split the list into the four quads
        mid_x_coord = self.nodes[0].bounding_box[2]
        mid_y_coord = self.nodes[0].bounding_box[1]

        cdef int left_y_split_index, x_split_index, right_y_split_index
        _sort_list_to_quads(
            point_list, left_bound, right_bound, mid_x_coord,
            mid_y_coord, &left_y_split_index, &x_split_index,
            &right_y_split_index)

        # quads indexed like this:
        # 01
        # 23

        if left_bound != left_y_split_index:
            self.nodes[2].add_points(
                point_list, left_bound, left_y_split_index)
        if left_y_split_index != x_split_index:
            self.nodes[0].add_points(
                point_list, left_y_split_index, x_split_index)
        if x_split_index != right_y_split_index:
            self.nodes[3].add_points(
                point_list, x_split_index, right_y_split_index)
        if right_y_split_index != right_bound:
            self.nodes[1].add_points(
                point_list, right_y_split_index, right_bound)

    def _bounding_box_intersect(self, bb):
        """Test if this node's bounding intersects another.

        Args:
            bb (list): bounding box of form [xmin, ymin, xmax, ymax]

        Returns:
            True if self.bounding_box intersects with bb, False otherwise
        """
        return not (
            self.bounding_box[0] > bb[2] or
            self.bounding_box[2] < bb[0] or
            self.bounding_box[1] > bb[3] or
            self.bounding_box[3] < bb[1])

    def n_nodes(self):
        """Return the number of nodes in the quadtree"""
        if self.is_leaf:
            return 1
        return sum([self.nodes[index].n_nodes() for index in xrange(4)]) + 1

    def n_points(self):
        """Return the number of nodes in the quadtree"""
        if self.is_leaf:
            return self.n_points_in_node
        return sum([self.nodes[index].n_points() for index in xrange(4)])

    def get_intersecting_points_in_polygon(self, shapely_polygon):
        """Return the points contained in `shapely_prepared_polygon`.

        This function is a high performance test routine to return the points
        contained in the shapely_prepared_polygon that are stored in `self`'s
        representation of a quadtree.

        Args:
            shapely_polygon (ogr.DataSource): a polygon datasource to bound
                against

        Returns:
            deque of (data, x_coord, y_coord) of nodes that are contained
                in `shapely_prepared_polygon`.
        """
        bounding_polygon = shapely.geometry.box(*self.bounding_box)

        if self.is_leaf:
            if shapely_polygon.contains(bounding_polygon):
                # trivial, all points are in the poly
                return self._get_points_from_node()
            elif shapely_polygon.intersects(bounding_polygon):
                # tricky, some points might be in poly
                result_deque = collections.deque()
                shapely_prepared_polygon = shapely.prepared.prep(
                    shapely_polygon)
                polygon_box = shapely.geometry.box(
                    *shapely_polygon.bounds)
                for point in self._get_points_from_node():
                    shapely_point = shapely.geometry.Point(point[2], point[3])
                    if (polygon_box.contains(shapely_point) and
                            shapely_prepared_polygon.contains(shapely_point)):
                        result_deque.append(point)
                return result_deque
        elif shapely_polygon.intersects(bounding_polygon):
            # combine results of children
            result_deque = collections.deque()
            for node_index in xrange(4):
                result_deque.extend(
                    self.nodes[node_index].get_intersecting_points_in_polygon(
                        shapely_polygon))
            return result_deque

        return collections.deque()  # empty

    def get_intersecting_points_in_bounding_box(self, bounding_box):
        """Get list of data that is contained by bounding_box.

        This function takes in a bounding box and returns a list of
        (data, lat, lng) tuples that are contained in the leaf nodes that
        intersect that bounding box.

        Args:
            bounding_box (list): of the form [xmin, ymin, xmax, ymax]

        Returns:
            numpy.ndarray array of (data, x_coord, lng) of nodes that
            intersect the bounding box.
        """
        if not self._bounding_box_intersect(bounding_box):
            return numpy.empty(0, dtype=_ARRAY_TUPLE_TYPE)

        if self.is_leaf:
            # drain the node into a list, filter to current bounding box
            point_list = numpy.array([
                point for point in self._get_points_from_node() if _in_box(
                    bounding_box, point[2], point[3])],
                dtype=_ARRAY_TUPLE_TYPE)
            return point_list
        else:
            point_list = numpy.empty(0, dtype=_ARRAY_TUPLE_TYPE)
            for node_index in xrange(4):
                point_list = numpy.concatenate((
                    self.nodes[node_index].get_intersecting_points_in_bounding_box(
                        bounding_box), point_list))
            return point_list


cdef _sort_list_to_quads(
        numpy.ndarray point_list, int left_bound, int right_bound,
        float mid_x_coord, float mid_y_coord, int *left_y_split_index,
        int *x_split_index, int *right_y_split_index):
    """Sort `point_list` points into quads.

        Sort the points in point_list to be 4 sublists where each sublists's
        points lie within the quads indexed like this:

        01
        23

        Modifies the values of `left_y_split_index`, `x_split_index`, and
        `right_y_split_index` to report split.

        Args:
            point_list (numpy.ndarray): structured numpy array of
                (data, x_coord, y_coord) points. This parameter will be
                modified to have sorted points
            left_bound (int): left inclusive index in `point_list` to sort
            right_bound (int): right non-inclusive index in `point_list` to sort
            list_bounds (tuple): (left, right) tuple of valid points in
                `point_list`
            mid_x_coord (float): x coord to split the quad
            mid_y_coord (float): y coord to split the quad

        Returns:
            None
    """
    # sort by x coordinates
    cdef numpy.ndarray sub_array = point_list[left_bound:right_bound]
    cdef numpy.ndarray idx = sub_array['f2'].argsort()

    sub_array[:] = sub_array[idx]
    x_split_index[0] = sub_array['f2'].searchsorted(mid_x_coord) + left_bound

    # sort the left y coordinates
    sub_array = point_list[left_bound:x_split_index[0]]
    idx = sub_array['f3'].argsort()
    sub_array[:] = sub_array[idx]
    left_y_split_index[0] = sub_array['f3'].searchsorted(
        mid_y_coord) + left_bound

    # sort the right y coordinates
    sub_array = point_list[x_split_index[0]:right_bound]
    idx = sub_array['f3'].argsort()
    sub_array[:] = sub_array[idx]
    right_y_split_index[0] = sub_array['f3'].searchsorted(
        mid_y_coord) + x_split_index[0]


def _in_box(bounding_box, x_coord, y_coord):
    """Test if coordinate is contained in the bounding box.

    Args:
        bounding_box (list): of form [xmin, ymin, xmax, ymax]
        x_coord, y_coord (int): x and y coordinate to test.

    Returns:
        True if (x_coord, y_coord) bound in bounding_box, False otherwise.
    """
    return (
        x_coord >= bounding_box[0] and  # xmin
        y_coord >= bounding_box[1] and  # ymin
        x_coord < bounding_box[2] and  # xmax
        y_coord < bounding_box[3])  # ymax



# coding=UTF-8

from natcap.invest.ui import model, inputs
from natcap.invest import coastal_vulnerability, MODEL_METADATA

from osgeo import gdal


class CoastalVulnerability(model.InVESTModel):
    def __init__(self):
        model.InVESTModel.__init__(
            self,
            label=MODEL_METADATA['coastal_vulnerability'].model_title,
            target=coastal_vulnerability.execute,
            validator=coastal_vulnerability.validate,
            localdoc=MODEL_METADATA['coastal_vulnerability'].userguide)

        self.aoi_vector_path = inputs.File(
            args_key='aoi_vector_path',
            helptext=(
                "Path to a polygon vector that is projected in "
                "a coordinate system with units of meters. "
                "Shore points will be created along all landmasses "
                "within the AOI polygon(s)."),
            label='Area of Interest (Vector)',
            validator=self.validator)
        self.add_input(self.aoi_vector_path)

        self.model_resolution = inputs.Text(
            args_key='model_resolution',
            helptext=(
                "Distance in meters between each shore point, "
                "as measured along the landmass polygon coastline."),
            label='Model resolution (meters)',
            validator=self.validator)
        self.add_input(self.model_resolution)

        self.landmass_vector_path = inputs.File(
            args_key='landmass_vector_path',
            helptext=(
                "Path to a polygon vector representing landmasses "
                "in the region of interest."),
            label='Landmass (Vector)',
            validator=self.validator)
        self.add_input(self.landmass_vector_path)

        self.wwiii_vector_path = inputs.File(
            args_key='wwiii_vector_path',
            helptext=(
                "Path to a point vector containing wind and wave "
                "data. This global dataset is provided with the InVEST "
                "sample data."),
            label='WaveWatchIII (Vector)',
            validator=self.validator)
        self.add_input(self.wwiii_vector_path)

        self.max_fetch_distance = inputs.Text(
            args_key='max_fetch_distance',
            helptext=(
                "Maximum distance in meters to extend rays from shore points. "
                "Rays extend in 16 compass directions until they intersect "
                "land or reach this maximum distance. Fetch (or ray) length "
                "is a component of wind and wave exposure."),
            label='Maximum Fetch Distance (meters)',
            validator=self.validator)
        self.add_input(self.max_fetch_distance)

        self.bathymetry_raster_path = inputs.File(
            args_key='bathymetry_raster_path',
            helptext=(
                "Path to a raster representing bathymetry. "
                "Bathymetry values should be negative and units of meters. "
                "Positive values will be ignored."),
            label='Bathymetry (Raster)',
            validator=self.validator)
        self.add_input(self.bathymetry_raster_path)

        self.dem_path = inputs.File(
            args_key='dem_path',
            helptext=(
                "Path to a raster representing elevation on land. "
                "Negative values, if present, are treated as zeros. "
                "If this input is unprojected, the model will project "
                "it to match the AOI and resample it to a cell size "
                "equal to the model resolution."),
            label='Digital Elevation Model (Raster)',
            validator=self.validator)
        self.add_input(self.dem_path)

        self.dem_averaging_radius = inputs.Text(
            args_key='dem_averaging_radius',
            helptext=(
                "A radius around each shore point within which to "
                "average the elevation values of the DEM raster. "),
            label='Elevation averaging radius (meters)',
            validator=self.validator)
        self.add_input(self.dem_averaging_radius)

        self.shelf_contour_vector_path = inputs.File(
            args_key='shelf_contour_vector_path',
            helptext=(
                "Path to a polyline vector delineating the edge "
                "of the continental shelf or another bathymetry contour. "),
            label='Continental Shelf Contour (Vector)',
            validator=self.validator)
        self.add_input(self.shelf_contour_vector_path)

        self.habitat_table_path = inputs.File(
            args_key='habitat_table_path',
            helptext=(
                "Path to a CSV file that specifies habitat layer input data "
                "and parameters."),
            label='Habitats Table (CSV)',
            validator=self.validator)
        self.add_input(self.habitat_table_path)

        self.geomorphology_vector_path = inputs.File(
            args_key='geomorphology_vector_path',
            helptext=(
                "Path to a polyline vector that has a field called "
                "'RANK' with values from 1 to 5. "),
            label='Geomorphology (Vector) (optional)',
            validator=self.validator)
        self.add_input(self.geomorphology_vector_path)

        self.geomorphology_fill_value = inputs.Dropdown(
            args_key='geomorphology_fill_value',
            helptext=(
                "A value from 1 to 5 that will be used as a geomorphology "
                "rank for any points not proximate (given the model "
                "resolution) to the geomorphology_vector_path."),
            label='Geomorphology fill value',
            interactive=False,
            options=('1', '2', '3', '4', '5'))
        self.add_input(self.geomorphology_fill_value)

        self.population_raster_path = inputs.File(
            args_key='population_raster_path',
            helptext=(
                "Path to a raster with values representing totals per pixel. "
                "If this input is unprojected, the model will project "
                "it to match the AOI and resample it to a cell size "
                "equal to the model resolution."),
            label='Human Population (Raster) (optional)',
            validator=self.validator)
        self.add_input(self.population_raster_path)

        self.population_radius = inputs.Text(
            args_key='population_radius',
            helptext=(
                "A radius around each shore point within which to "
                "compute the average population density."),
            label='Population search radius (meters)',
            interactive=False,
            validator=self.validator)
        self.add_input(self.population_radius)

        self.slr_vector_path = inputs.File(
            args_key='slr_vector_path',
            helptext=(
                "Path to a point vector with a field of sea-level-rise rates"
                "or amounts."),
            label='Sea Level Rise (Vector) (optional)',
            validator=self.validator)
        self.add_input(self.slr_vector_path)

        self.slr_field = inputs.Dropdown(
            args_key='slr_field',
            helptext=(
                "The name of a field in the SLR vector table"
                "from which to load values"),
            label='Sea Level Rise fieldname',
            interactive=False,
            options=('UNKNOWN',))  # No options until valid OGR vector provided
        self.add_input(self.slr_field)

        # Set interactivity requirement as input sufficiency changes
        self.slr_vector_path.sufficiency_changed.connect(
            self.slr_field.set_interactive)
        self.slr_vector_path.sufficiency_changed.connect(
            self._load_colnames)
        self.geomorphology_vector_path.sufficiency_changed.connect(
            self.geomorphology_fill_value.set_interactive)
        self.population_raster_path.sufficiency_changed.connect(
            self.population_radius.set_interactive)

    def _load_colnames(self, new_interactivity):
        if new_interactivity:
            new_vector_path = self.slr_vector_path.value()
            new_vector = gdal.OpenEx(new_vector_path, gdal.OF_VECTOR)
            if new_vector is not None:
                new_layer = new_vector.GetLayer()
                colnames = [defn.GetName() for defn in new_layer.schema]
                self.slr_field.set_options(colnames)
                self.slr_field.set_interactive(True)
            else:
                self.slr_field.set_options([])
                self.slr_field.set_interactive(False)
        elif not self.slr_vector_path.sufficient:
            self.slr_field.set_options([])

    def assemble_args(self):
        args = {
            self.workspace.args_key: self.workspace.value(),
            self.suffix.args_key: self.suffix.value(),
            self.aoi_vector_path.args_key: self.aoi_vector_path.value(),
            self.model_resolution.args_key:
                self.model_resolution.value(),
            self.landmass_vector_path.args_key:
                self.landmass_vector_path.value(),
            self.wwiii_vector_path.args_key:
                self.wwiii_vector_path.value(),
            self.max_fetch_distance.args_key: self.max_fetch_distance.value(),
            self.bathymetry_raster_path.args_key:
                self.bathymetry_raster_path.value(),
            self.dem_path.args_key: self.dem_path.value(),
            self.dem_averaging_radius.args_key:
                self.dem_averaging_radius.value(),
            self.shelf_contour_vector_path.args_key:
                self.shelf_contour_vector_path.value(),
            self.habitat_table_path.args_key:
                self.habitat_table_path.value(),
            self.geomorphology_vector_path.args_key:
                self.geomorphology_vector_path.value(),
            self.geomorphology_fill_value.args_key:
                self.geomorphology_fill_value.value(),
            self.population_raster_path.args_key:
                self.population_raster_path.value(),
            self.population_radius.args_key:
                self.population_radius.value(),
            self.slr_vector_path.args_key:
                self.slr_vector_path.value(),
            self.slr_field.args_key:
                self.slr_field.value(),
        }

        return args



"""A class for a Qt-enabled python Thread."""
import threading
import logging
import traceback

from qtpy import QtCore


LOGGER = logging.getLogger(__name__)
LOGGER.addHandler(logging.NullHandler())


class Executor(QtCore.QObject, threading.Thread):
    """A thread of control that will emit a Qt signal when finished."""

    finished = QtCore.Signal()

    def __init__(self, target, args=None, kwargs=None, log_events=True):
        """Initialize the Executor object.

        Args:
            target (callable): A function or unbound method that should be
                called within the separate thread of control.
            args (iterable): An iterable of positional arguments that will
                be passed to ``target``.  If ``None``, positional
                arguments will be ignored.
            kwargs (dict): A dict mapping string parameter names to parameter
                values that should be passed to ``target``.  If ``None``,
                keyword arguments will be ignored.
            log_events=True (bool): If ``True``, exceptions raised when calling
                ``target`` as well as completion of ``target`` will
                be logged.

        Returns:
            ``None``.
        """
        QtCore.QObject.__init__(self)
        threading.Thread.__init__(self)
        self.target = target
        self.log_events = log_events

        if args is None:
            args = ()
        self.args = args

        if kwargs is None:
            kwargs = {}
        self.kwargs = kwargs

        self.failed = False
        self.exception = None
        self.traceback = None

    def run(self):
        """Run the target callable in a separate thread of control.

        The callable will be run with whatever ``args`` and ``kwargs`` are
        provided to the thread's ``__init__`` method.

        If an exception is encountered while executing the target, several
        things happen:

            * The exception is logged.
            * ``self.failed`` is set to ``True``.
            * ``self.exception`` refers to the exception object that was raised.
            * ``self.traceback`` refers to the formatted traceback.

        Finally, the signal ``self.finished`` is emitted, regardless of whether
        an exception was raised.
        """
        try:
            self.target(*self.args, **self.kwargs)
        except BaseException as error:
            # We deliberately want to catch all possible exceptions, so
            # BaseException is the way to go.  This is in part because we have
            # a flaky test that failed where self.exception wasn't set and we
            # don't know if that's because this exception handler wasn't being
            # called (we were only capturing Exception at the time), or if
            # there was something else very strange going on.
            # Failed build:
            # http://builds.naturalcapitalproject.org/job/test-natcap.invest.ui/100/

            # When we're running a model, the exception is logged via
            # utils.prepare_workspace.  But this thread is also used in the Qt
            # interface by validation and the datastack archive creation
            # function, and we for sure want to log exceptions in that case.
            if self.log_events:
                LOGGER.exception('Target %s failed with exception', self.target)
            self.failed = True
            self.exception = error
            self.traceback = traceback.format_exc()
        finally:
            if self.log_events:
                LOGGER.info('Execution finished')
            self.finished.emit()



"""A Flask app with HTTP endpoints used by the InVEST Workbench."""
import importlib
import json
import logging

from osgeo import gdal
from flask import Flask
from flask import request
import natcap.invest
from natcap.invest import cli
from natcap.invest import datastack
from natcap.invest import install_language
from natcap.invest import MODEL_METADATA
from natcap.invest import spec_utils
from natcap.invest import usage

LOGGER = logging.getLogger(__name__)

app = Flask(__name__)

PYNAME_TO_MODEL_NAME_MAP = {
    metadata.pyname: model_name
    for model_name, metadata in MODEL_METADATA.items()
}


def shutdown_server():
    """Shutdown the flask server."""
    func = request.environ.get('werkzeug.server.shutdown')
    if func is None:
        raise RuntimeError('Not running with the Werkzeug Server')
    func()


@app.route('/ready', methods=['GET'])
def get_is_ready():
    """Returns something simple to confirm the server is open."""
    return 'Flask ready'


@app.route('/shutdown', methods=['GET'])
def shutdown():
    """A request to this endpoint shuts down the server."""
    shutdown_server()
    return 'Flask server shutting down...'


@app.route('/models', methods=['GET'])
def get_invest_models():
    """Gets a list of available InVEST models.

    Accepts a `language` query parameter which should be an ISO 639-1 language
    code. Model names will be translated to the requested language if
    translations are available, or fall back to English otherwise.

    Returns:
        A JSON string
    """
    LOGGER.debug('get model list')
    install_language(request.args.get('language', 'en'))
    importlib.reload(natcap.invest)
    return cli.build_model_list_json()


@app.route('/getspec', methods=['POST'])
def get_invest_getspec():
    """Gets the ARGS_SPEC dict from an InVEST model.

    Body (JSON string): "carbon"
    Accepts a `language` query parameter which should be an ISO 639-1 language
    code. Spec 'about' and 'name' values will be translated to the requested
    language if translations are available, or fall back to English otherwise.

    Returns:
        A JSON string.
    """
    install_language(request.args.get('language', 'en'))
    target_model = request.get_json()
    target_module = MODEL_METADATA[target_model].pyname
    model_module = importlib.reload(
        importlib.import_module(name=target_module))
    return spec_utils.serialize_args_spec(model_module.ARGS_SPEC)


@app.route('/validate', methods=['POST'])
def get_invest_validate():
    """Gets the return value of an InVEST model's validate function.

    Body (JSON string):
        model_module: string (e.g. natcap.invest.carbon)
        args: JSON string of InVEST model args keys and values

    Accepts a `language` query parameter which should be an ISO 639-1 language
    code. Validation messages will be translated to the requested language if
    translations are available, or fall back to English otherwise.

    Returns:
        A JSON string.
    """
    payload = request.get_json()
    LOGGER.debug(payload)
    try:
        limit_to = payload['limit_to']
    except KeyError:
        limit_to = None

    install_language(request.args.get('language', 'en'))
    importlib.reload(natcap.invest.validation)
    model_module = importlib.reload(
        importlib.import_module(name=payload['model_module']))

    results = model_module.validate(
        json.loads(payload['args']), limit_to=limit_to)
    LOGGER.debug(results)
    return json.dumps(results)


@app.route('/colnames', methods=['POST'])
def get_vector_colnames():
    """Get a list of column names from a vector.
    This is used to fill in dropdown menu options in a couple models.

    Body (JSON string):
        vector_path (string): path to a vector file

    Returns:
        a JSON string.
    """
    payload = request.get_json()
    LOGGER.debug(payload)
    vector_path = payload['vector_path']
    # a lot of times the path will be empty so don't even try to open it
    if vector_path:
        try:
            vector = gdal.OpenEx(vector_path, gdal.OF_VECTOR)
            colnames = [defn.GetName() for defn in vector.GetLayer().schema]
            LOGGER.debug(colnames)
            return json.dumps(colnames)
        except Exception as e:
            LOGGER.exception(
                f'Could not read column names from {vector_path}. ERROR: {e}')
    else:
        LOGGER.error('Empty vector path.')
    # 422 Unprocessable Entity: the server understands the content type
    # of the request entity, and the syntax of the request entity is
    # correct, but it was unable to process the contained instructions.
    return json.dumps([]), 422


@app.route('/post_datastack_file', methods=['POST'])
def post_datastack_file():
    """Extracts InVEST model args from json, logfiles, or datastacks.

    Body (JSON string): path to file

    Returns:
        A JSON string.
    """
    filepath = request.get_json()
    stack_type, stack_info = datastack.get_datastack_info(
        filepath)
    model_name = PYNAME_TO_MODEL_NAME_MAP[stack_info.model_name]
    result_dict = {
        'type': stack_type,
        'args': stack_info.args,
        'module_name': stack_info.model_name,
        'model_run_name': model_name,
        'model_human_name': MODEL_METADATA[model_name].model_title,
        'invest_version': stack_info.invest_version
    }
    return json.dumps(result_dict)


@app.route('/write_parameter_set_file', methods=['POST'])
def write_parameter_set_file():
    """Writes InVEST model args keys and values to a datastack JSON file.

    Body (JSON string):
        parameterSetPath: string
        moduleName: string(e.g. natcap.invest.carbon)
        args: JSON string of InVEST model args keys and values
        relativePaths: boolean

    Returns:
        A string.
    """
    payload = request.get_json()
    filepath = payload['parameterSetPath']
    modulename = payload['moduleName']
    args = json.loads(payload['args'])
    relative_paths = payload['relativePaths']

    datastack.build_parameter_set(
        args, modulename, filepath, relative=relative_paths)
    return 'parameter set saved'


@app.route('/save_to_python', methods=['POST'])
def save_to_python():
    """Writes a python script with a call to an InVEST model execute function.

    Body (JSON string):
        filepath: string
        modelname: string (e.g. carbon)
        args_dict: JSON string of InVEST model args keys and values

    Returns:
        A string.
    """
    payload = request.get_json()
    save_filepath = payload['filepath']
    modelname = payload['modelname']
    args_dict = json.loads(payload['args'])

    cli.export_to_python(
        save_filepath, modelname, args_dict)

    return 'python script saved'


@app.route('/log_model_start', methods=['POST'])
def log_model_start():
    payload = request.get_json()
    usage._log_model(
        payload['model_pyname'],
        json.loads(payload['model_args']),
        payload['invest_interface'],
        payload['session_id'])
    return 'OK'


@app.route('/log_model_exit', methods=['POST'])
def log_model_exit():
    payload = request.get_json()
    usage._log_exit_status(
        payload['session_id'],
        payload['status'])
    return 'OK'
